"""Functional approach to Quantities."""

__all__: tuple[str, ...] = ()

from dataclasses import replace
from typing import Any

import equinox as eqx
import numpy as np
from astropy.units import UnitConversionError
from jaxtyping import Array, ArrayLike
from plum import dispatch

import unxt_api as uapi
from .base import AbstractQuantity
from .base_angle import AbstractAngle
from .base_parametric import AbstractParametricQuantity
from .static_quantity import StaticQuantity
from .value import StaticValue
from unxt.dims import AbstractDimension
from unxt.units import AbstractUnit
from unxt.unitsystems import AbstractUnitSystem

# ===================================================================
# Get dimensions


@dispatch
def dimension_of(obj: AbstractQuantity, /) -> AbstractDimension:
    """Return the dimension of a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> q = u.Q(1, "m")
    >>> u.dimension_of(q)
    PhysicalType('length')

    """
    return uapi.dimension_of(obj.unit)


@dispatch
def dimension_of(obj: type[AbstractParametricQuantity], /) -> AbstractDimension:
    """Return the dimension of a quantity.

    Examples
    --------
    >>> import unxt as u

    >>> try:
    ...     u.dimension_of(u.Quantity)
    ... except Exception as e:
    ...     print(e)
    can only get dimensions from parametrized Quantity -- Quantity[dim].

    >>> u.dimension_of(u.Quantity["length"])
    PhysicalType('length')

    """
    obj = eqx.error_if(
        obj,
        not hasattr(obj, "_type_parameter"),
        f"can only get dimensions from parametrized {obj.__name__} "
        f"-- {obj.__name__}[dim].",
    )
    return obj._type_parameter  # noqa: SLF001


@dispatch
def dimension_of(obj: type[AbstractAngle], /) -> AbstractDimension:
    """Get the dimension of an angle class.

    Examples
    --------
    >>> import unxt as u

    >>> u.dimension_of(u.Angle)
    PhysicalType('angle')

    """
    return uapi.dimension("angle")


# ===================================================================
# Get units


@dispatch
def unit_of(obj: AbstractQuantity, /) -> AbstractUnit:
    """Return the units of an object.

    Examples
    --------
    >>> from unxt import unit_of, Quantity
    >>> q = Quantity(1, "m")
    >>> unit_of(q)
    Unit("m")

    """
    return obj.unit


# ===================================================================
# Convert units for values


@dispatch
def uconvert_value(
    tousys: AbstractUnitSystem, ufrom: AbstractUnit, x: ArrayLike, /
) -> ArrayLike:
    """Convert the value from units to a unitsystem's preferred units.

    Examples
    --------
    >>> import unxt as u
    >>> u.uconvert_value(u.unitsystems.galactic, u.unit("km"), 1e17)  # kpc
    3.2407792894443648

    >>> u.unitsystems.galactic[u.dimension("length")]  # checking the units
    Unit("kpc")

    """
    return uapi.uconvert_value(tousys[uapi.dimension_of(ufrom)], ufrom, x)


@dispatch
def uconvert_value(
    tousys: AbstractUnitSystem, ufrom: str, x: ArrayLike, /
) -> ArrayLike:
    """Convert the value from units to a unitsystem's preferred units.

    Examples
    --------
    >>> import unxt as u
    >>> u.uconvert_value(u.unitsystems.galactic, "km", 1e17)  # in kpc
    3.2407792894443648

    >>> u.unitsystems.galactic[u.dimension("length")]  # checking the units
    Unit("kpc")

    """
    return uapi.uconvert_value(tousys, uapi.unit(ufrom), x)


@dispatch
def uconvert_value(uto: str, ufrom: str, x: ArrayLike, /) -> ArrayLike:
    """Convert the value to the specified units.

    Examples
    --------
    >>> import unxt as u

    >>> u.uconvert_value("m", "km", 1)
    1000.0

    """
    return uapi.uconvert_value(uapi.unit(uto), uapi.unit(ufrom), x)


# **NOTE:** we also add convenience dispatches for `uconvert_value` so that
# users can use the lower-level function with Quantity and not break their code.


@dispatch
def uconvert_value(uto: Any, ufrom: Any, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    This is a convenience dispatch so that users can use the lower-level
    function with Quantity and not break their code.
    This dispatch simply calls `uconvert`, checking first that the units are
    convertible.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Q(1, "km")
    >>> u.uconvert_value("m", "km", q)
    Quantity(Array(1000., dtype=float32, ...), unit='m')

    """
    x = eqx.error_if(
        x,
        not uapi.is_unit_convertible(ufrom, x.unit),
        f"Cannot convert from {x.unit} to {ufrom}",
    )
    return uapi.uconvert(uto, x)


# ===================================================================
# Convert units for quantities


@dispatch
def uconvert(ustr: str, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> from unxt import Quantity, units

    >>> x = Quantity(1000, "m")
    >>> uconvert("km", x)
    Quantity(Array(1., dtype=float32, ...), unit='km')

    """
    return uconvert(uapi.unit(ustr), x)


@dispatch
def uconvert(usys: AbstractUnitSystem, x: AbstractQuantity, /) -> AbstractQuantity:
    """Convert the quantity to the specified units.

    Examples
    --------
    >>> from unxt import Quantity, units
    >>> from unxt.unitsystems import galactic

    >>> q = Quantity(1e17, "km")
    >>> uconvert(galactic, q)
    Quantity(Array(3.2407792, dtype=float32, ...), unit='kpc')

    """
    return uconvert(usys[uapi.dimension_of(x)], x)


# ===================================================================
# Strip units


@dispatch
def ustrip(x: StaticQuantity, /) -> np.ndarray:
    """Strip the units from a static quantity."""
    return x.value.array


@dispatch
def ustrip(x: AbstractQuantity, /) -> Array | np.ndarray:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Q(1000, "m")
    >>> u.ustrip(q)
    Array(1000, dtype=int32, weak_type=True)

    >>> u.ustrip(q) is q.value
    True

    """
    v = x.value
    return v.array if isinstance(v, StaticValue) else v


@dispatch
def ustrip(u: AbstractUnit, x: AbstractQuantity, /) -> Array | np.ndarray:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Q(1000, "m")
    >>> u.ustrip(u.unit("km"), q)
    Array(1., dtype=float32, ...)

    """
    return uapi.ustrip(uapi.uconvert(u, x))


@dispatch
def ustrip(u: str, x: AbstractQuantity, /) -> Array | np.ndarray:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u

    >>> q = u.Q(1000, "m")
    >>> u.ustrip("km", q)
    Array(1., dtype=float32, ...)

    """
    return uapi.ustrip(uapi.unit(u), x)


@dispatch
def ustrip(u: AbstractUnitSystem, x: AbstractQuantity, /) -> Array | np.ndarray:
    """Strip the units from the quantity.

    Examples
    --------
    >>> import unxt as u
    >>> from unxt.unitsystems import galactic

    >>> q = u.Q(1e17, "km")
    >>> u.ustrip(galactic, q)
    Array(3.2407792, dtype=float32, weak_type=True)

    """
    return uapi.ustrip(u[uapi.dimension_of(x)], x)


# ============================================================================


@dispatch
def is_unit_convertible(to_unit: Any, from_: Any, /) -> bool:
    """Check if a unit can be converted to another unit.

    Parameters
    ----------
    to_unit : Any
        The unit to convert to. Converted to a unit object using `unxt.unit`.
    from_ : Any
        The unit to convert from. Converted to a unit object using `unxt.unit`,
        Note this means it also support `Quantity` objects and many others.

    Examples
    --------
    >>> from unxt import is_unit_convertible
    >>> is_unit_convertible("cm", "m")
    True

    >>> is_unit_convertible("m", "Gyr")
    False

    """
    to_u = uapi.unit(to_unit)
    from_u = uapi.unit(from_)
    try:
        from_u.to(to_u)
    except UnitConversionError:
        return False
    return True


@dispatch
def is_unit_convertible(to_unit: Any, from_: AbstractQuantity, /) -> bool:
    """Check if a Quantity can be converted to another unit.

    Examples
    --------
    >>> from unxt import Quantity, is_unit_convertible
    >>> q = Quantity(1, "m")

    >>> is_unit_convertible("cm", q)
    True

    >>> is_unit_convertible("Gyr", q)
    False

    """
    return is_unit_convertible(to_unit, from_.unit)


# ============================================================================


@dispatch
def wrap_to(
    angle: AbstractQuantity,
    min: AbstractQuantity,
    max: AbstractQuantity,
    /,
) -> AbstractQuantity:
    """Wrap to the range [min, max).

    Examples
    --------
    >>> import unxt as u

    >>> angle = u.Angle(370, "deg")
    >>> u.quantity.wrap_to(angle, min=u.Q(0, "deg"), max=u.Q(360, "deg"))
    Angle(Array(10, dtype=int32, ...), unit='deg')

    """
    minv, maxv = min.ustrip(angle.unit), max.ustrip(angle.unit)
    value = ((angle.value - minv) % (maxv - minv)) + minv
    return replace(angle, value=value)



"""Doctest configuration."""

import os
from collections.abc import Callable, Iterable, Sequence
from doctest import ELLIPSIS, NORMALIZE_WHITESPACE
from typing import Any

from sybil import Document, Region, Sybil
from sybil.parsers import myst, rest

from optional_dependencies import OptionalDependencyEnum, auto

optionflags = ELLIPSIS | NORMALIZE_WHITESPACE

parsers: Sequence[Callable[[Document], Iterable[Region]]] = [
    myst.DocTestDirectiveParser(optionflags=optionflags),
    myst.PythonCodeBlockParser(doctest_optionflags=optionflags),
    myst.SkipParser(),
]

docs = Sybil(parsers=parsers, patterns=["*.md"])
python = Sybil(
    parsers=[
        *parsers,
        rest.PythonCodeBlockParser(),
        rest.DocTestParser(optionflags=optionflags),
        rest.SkipParser(),
    ],
    patterns=["*.py"],
)


pytest_collect_file = (docs + python).pytest()


class OptDeps(OptionalDependencyEnum):
    """Optional dependencies for ``unxt``."""

    ASTROPY = auto()
    GALA = auto()
    MATPLOTLIB = auto()


collect_ignore_glob = []
if not OptDeps.ASTROPY.installed:
    collect_ignore_glob.append("src/unxt/_interop/unxt_interop_astropy/*")
if not OptDeps.GALA.installed:
    collect_ignore_glob.append("src/unxt/_interop/unxt_interop_gala/*")
    collect_ignore_glob.append("docs/interop/gala.md")
if not OptDeps.MATPLOTLIB.installed:
    collect_ignore_glob.append("src/unxt/_interop/unxt_interop_mpl/*")
    collect_ignore_glob.append("docs/interop/matplotlib.md")


def pytest_generate_tests(metafunc: Any) -> None:
    os.environ["UNXT_ENABLE_RUNTIME_TYPECHECKING"] = "beartype.beartype"



"""Units API for unxt.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__ = ("unit", "unit_of")

from typing import Any

import plum


@plum.dispatch.abstract
def unit(obj: Any, /) -> Any:
    """Construct the units from a units object."""


@plum.dispatch.abstract
def unit_of(obj: Any, /) -> Any:
    """Return the units of an object."""



# pylint: disable=import-error, no-member, unsubscriptable-object
#    b/c it doesn't understand dataclass fields

__all__ = ("AbstractParametricQuantity",)

from collections.abc import Callable
from functools import partial
from typing import Any

import equinox as eqx
import jax
import wadler_lindig as wl
from astropy.units import PhysicalType
from jaxtyping import Array, Shaped
from plum import dispatch, parametric, type_unparametrized

from dataclassish import field_items, fields

from .base import AbstractQuantity
from .value import StaticValue
from unxt._src.dimensions import name_of
from unxt.dims import AbstractDimension, dimension, dimension_of
from unxt.units import AbstractUnit, unit as parse_unit


@parametric
class AbstractParametricQuantity(AbstractQuantity):
    """Arrays with associated units.

    This class is parametrized by the dimensions of the units.

    """

    value: eqx.AbstractVar[Shaped[Array | StaticValue, "*shape"]]
    """The value of the `AbstractQuantity`."""

    unit: eqx.AbstractVar[AbstractUnit]
    """The unit associated with this value."""

    def __post_init__(self) -> None:
        """Check whether the arguments are valid."""
        self._type_parameter: AbstractDimension

    def __check_init__(self) -> None:
        """Check whether the arguments are valid."""
        expected_dimensions = self._type_parameter
        got_dimensions = dimension_of(self.unit)
        if got_dimensions != expected_dimensions:  # pylint: disable=unreachable
            msg = "Physical type mismatch."  # TODO: better error message
            raise ValueError(msg)

    # ---------------------------------------------------------------
    # Plum stuff

    @classmethod
    @dispatch
    def __init_type_parameter__(
        cls, dims: AbstractDimension, /
    ) -> tuple[AbstractDimension]:
        """Check whether the type parameters are valid."""
        return (dims,)

    @classmethod
    @dispatch
    def __init_type_parameter__(cls, dims: str, /) -> tuple[AbstractDimension]:
        """Check whether the type parameters are valid."""
        dims_: AbstractDimension
        try:
            dims_ = dimension(dims)
        except ValueError:
            dims_ = dimension_of(parse_unit(dims))
        return (dims_,)

    @classmethod
    @dispatch
    def __init_type_parameter__(cls, unit: AbstractUnit, /) -> tuple[AbstractDimension]:
        """Infer the type parameter from the arguments."""
        dims = dimension_of(unit)
        if dims != "unknown":  # pylint: disable=unreachable
            return (dims,)
        return (PhysicalType(unit, unit.to_string(fraction=False)),)

    @classmethod
    def __infer_type_parameter__(
        cls, value: Any, unit: Any, **kwargs: Any
    ) -> tuple[AbstractDimension]:
        """Infer the type parameter from the arguments."""
        return (dimension_of(parse_unit(unit)),)

    @classmethod
    @dispatch
    def __le_type_parameter__(
        cls, left: tuple[AbstractDimension], right: tuple[AbstractDimension]
    ) -> bool:
        """Define an order on type parameters.

        That is, check whether `left <= right` or not.
        """
        (dim_left,) = left
        (dim_right,) = right
        return dim_left == dim_right

    # ---------------------------------------------------------------
    # misc

    def __getnewargs_ex__(self) -> tuple[tuple[Any, ...], dict[str, Any]]:
        """Return args, kwargs for ``__new__``.

        In protocols 2+, this is used to determine the values (args,
        and kwargs) passed to ``__new__``. We implement ``__getnewargs_ex__``
        instead of ``__getnewargs__`` because the latter does not support
        keyword-only arguments.

        Examples
        --------
        >>> import copy as pycopy
        >>> import unxt as u

        >>> x = u.Quantity([1, 2, 3], "m")
        >>> pycopy.copy(x)
        Quantity(Array([1, 2, 3], dtype=int32), unit='m')

        """
        return (), field_items(self)

    # TODO: see if pickling can be accomplished without reduce.
    # https://docs.python.org/3.12/library/pickle.html
    def __reduce__(
        self,
    ) -> tuple[Callable[..., "AbstractParametricQuantity"], tuple[Any, ...]]:
        r"""Return the reduced value: a constructor and arguments.

        The ``__reduce__`` protocol has limited support for keyword-only
        argument. The only built-in means to pass kwargs is to the
        ``__setstate__`` method or through a callable with signature ``(obj,
        state)``. Neither of these methods allow for the kwargs to be passed to
        the constructor, only after the object is partially initialized. This
        does not work well for JAX, Equinox, or particularly run-time
        typechecking. To get around this, instead of the class type as the
        customary first element of reduced value we will instead return a
        `functools.partial`-wrapping of the class type with the kwargs bundled
        into the partial object. This is allowed by the ``__reduce__`` protocol,
        which says the first returned item must be any "callable object that
        will be called to create the initial version of the object." In
        conjunction with the standard args, the `functools.partial`-wrapped
        class type will properly construct any parametric subclass.

        Returns
        -------
        functools.partial[type]
            The `plum.type_unparametrized` form of this class object, which is
            the parametric class without the specialization to the specific
            parameter. The specialized subtype cannot be pickled since it is
            dynamically produced. The type is wrapped into a `functools.partial`
            with the keyword-only arguments to the constructor.
        tuple[Any, ...]
            The arguments to this class. Note: the keyword-only arguments are
            bundled with the type.

        Examples
        --------
        >>> import pickle
        >>> import unxt as u

        >>> x = u.Quantity([1, 2, 3], "m")
        >>> pickle.dumps(x)
        b'...'

        """
        args, kwargs = [], {}
        for f in fields(self):
            v = getattr(self, f.name)
            if f.kw_only:
                kwargs[f.name] = v
            else:
                args.append(v)

        return partial(type_unparametrized(self), **kwargs), tuple(args)

    def __pdoc__(
        self,
        *,
        include_params: bool = False,
        named_unit: bool = True,
        use_short_name: bool = False,
        **kwargs: Any,
    ) -> wl.AbstractDoc:
        """Return the Wadler-Lindig representation of this class.

        This is used for the `__repr__` and `__str__` methods or when using the
        `wadler_lindig` library.

        Parameters
        ----------
        include_params
            If `True`, the type parameter is included in the representation. If
            `False`, the type parameter is not included in the representation.
            For example, ``Quantity['length'][i32]`` versus ``Quantity[i32]``.
        named_unit
            If `True`, the unit is included in the representation as a named
            argument. If `False`, the unit is included as a positional argument.
            For example, ``Quantity(<array>, unit='m')`` versus
            ``Quantity(<array>, 'm')``.
        use_short_name
            If `True` and the class has a ``short_name`` class variable, use the
            short name instead of the full class name. For example, ``Q(...)``
            instead of ``Quantity(...)``.
        kwargs
            Additional keyword arguments ``wadler_lindig.pdoc`` method for
            formatting the value, stringified unit, and other fields.

        Examples
        --------
        >>> import unxt as u
        >>> import wadler_lindig as wl

        >>> q = u.Quantity([1, 2, 3], "m")

        The default pretty printing:

        >>> wl.pprint(q)
        Quantity(i32[3], unit='m')

        The type parameter can be included in the representation:

        >>> wl.pprint(q, include_params=True)
        Quantity['length'](i32[3], unit='m')

        The `str` method uses this as well:

        >>> print(q)
        Quantity['length']([1, 2, 3], unit='m')

        Arrays can be printed in full:

        >>> wl.pprint(q, short_arrays=False)
        Quantity(Array([1, 2, 3], dtype=int32), unit='m')

        The `repr` method uses this setting:

        >>> print(repr(q))
        Quantity(Array([1, 2, 3], dtype=int32), unit='m')

        The units can be turned from a named argument to a positional argument
        by setting `named_unit=False`:

        >>> wl.pprint(q, named_unit=False)
        Quantity(i32[3], 'm')

        The class can be printed with its short name:

        >>> wl.pprint(q, use_short_name=True)
        Q(i32[3], unit='m')

        Short name with type parameter:

        >>> wl.pprint(q, use_short_name=True, include_params=True)
        Q['length'](i32[3], unit='m')

        """
        pdoc = super().__pdoc__(
            include_params=include_params,
            named_unit=named_unit,
            use_short_name=use_short_name,
            **kwargs,
        )

        # Type Parameter
        if not include_params:
            param = wl.TextDoc("")
        else:
            param = wl.TextDoc(f"[{name_of(self._type_parameter)!r}]")

        return wl.ConcatDoc(pdoc.children[0], param, *pdoc.children[1:])

    def __repr__(self) -> str:
        return wl.pformat(
            self, include_params=False, named_unit=True, short_arrays=False, indent=4
        )

    def __str__(self) -> str:
        return wl.pformat(
            self,
            include_params=True,
            named_unit=True,
            short_arrays="compact"
            if not isinstance(self.value, jax.core.Tracer)
            else True,
            indent=4,
        )



"""Tests for the quantities strategy."""
# pylint: disable=unreachable

import hypothesis.strategies as st
import jax.numpy as jnp
from hypothesis import given, settings

import unxt as u
import unxt_hypothesis as ust


@given(q=ust.quantities(unit="m"))
@settings(max_examples=50)
def test_scalar_length(q: u.Quantity) -> None:
    """Test scalar length quantity generation."""
    assert isinstance(q, u.Quantity)
    assert q.shape == ()
    assert u.dimension_of(q) == u.dimension("length")


@given(q=ust.quantities(unit="m/s", shape=3))
@settings(max_examples=50)
def test_vector_velocity(q: u.Quantity) -> None:
    """Test 1D velocity quantity generation."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    assert u.dimension_of(q) == u.dimension("velocity")


@given(q=ust.quantities(unit="kg", shape=(2, 3)))
@settings(max_examples=50)
def test_2d_mass(q: u.Quantity) -> None:
    """Test 2D mass quantity generation."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (2, 3)
    assert u.dimension_of(q) == u.dimension("mass")


@given(q=ust.quantities(unit="J", dtype=jnp.float32))
@settings(max_examples=30)
def test_float32_energy_dtype(q: u.Quantity) -> None:
    """Test float32 dtype for energy."""
    assert isinstance(q, u.Quantity)
    assert q.dtype == jnp.float32
    assert u.dimension_of(q) == u.dimension("energy")


@given(q=ust.quantities(unit="s", dtype=jnp.float32, shape=5))
@settings(max_examples=30)
def test_float32_time_dtype(q: u.Quantity) -> None:
    """Test float32 dtype for time."""
    assert isinstance(q, u.Quantity)
    assert q.dtype == jnp.float32
    assert q.shape == (5,)
    assert u.dimension_of(q) == u.dimension("time")


@given(
    q=ust.quantities(unit="m", shape=st.tuples(st.integers(1, 5), st.integers(1, 5)))
)
@settings(max_examples=30)
def test_variable_shape_strategy(q: u.Quantity) -> None:
    """Test with a shape strategy."""
    assert isinstance(q, u.Quantity)
    assert len(q.shape) == 2
    assert all(1 <= s <= 5 for s in q.shape)
    assert u.dimension_of(q) == u.dimension("length")


@given(
    q=ust.quantities(
        unit="m",
        shape=3,
        elements=st.floats(min_value=0.0, max_value=100.0, width=32),
    )
)
@settings(max_examples=30, deadline=None)
def test_custom_elements(q: u.Quantity) -> None:
    """Test with custom element strategy."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    # All values should be between 0 and 100
    assert jnp.all((q.value >= 0) & (q.value <= 100))


@given(q=ust.quantities(unit="m", shape=10, unique=True))
@settings(max_examples=20, deadline=None)
def test_unique_elements(q: u.Quantity) -> None:
    """Test unique elements constraint."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (10,)
    # All elements should be unique
    assert len(jnp.unique(q.value)) == 10


@given(q=ust.quantities(unit="m", static_value=True))
@settings(max_examples=30)
def test_static_value_generation(q: u.Quantity) -> None:
    """Test StaticValue wrapping for generated quantities."""
    assert isinstance(q, u.Quantity)
    assert isinstance(q.value, u.quantity.StaticValue)


@given(q=ust.quantities(unit="rad"))
@settings(max_examples=30)
def test_angle_unit(q: u.Quantity) -> None:
    """Test with angle unit."""
    assert isinstance(q, u.Quantity)
    assert u.dimension_of(q) == u.dimension("angle")


@given(q=ust.quantities(unit="N"))
@settings(max_examples=30)
def test_force_unit(q: u.Quantity) -> None:
    """Test with force unit."""
    assert isinstance(q, u.Quantity)
    assert u.dimension_of(q) == u.dimension("force")


@given(q=ust.quantities(unit="Pa"))
@settings(max_examples=30)
def test_pressure_unit(q: u.Quantity) -> None:
    """Test with pressure unit."""
    assert isinstance(q, u.Quantity)
    assert u.dimension_of(q) == u.dimension("pressure")


@given(q=ust.quantities(unit="kpc"))
@settings(max_examples=30)
def test_astronomical_unit(q: u.Quantity) -> None:
    """Test with astronomical unit."""
    assert isinstance(q, u.Quantity)
    assert u.dimension_of(q) == u.dimension("length")


@given(q=ust.quantities(unit="km/s", shape=(3,)))
@settings(max_examples=30)
def test_compound_unit(q: u.Quantity) -> None:
    """Test with compound unit."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    assert u.dimension_of(q) == u.dimension("velocity")


def test_accepts_shape_none() -> None:
    """Test that shape=None creates scalars."""
    strategy = ust.quantities(unit="m", shape=None)
    assert strategy is not None


def test_accepts_shape_int() -> None:
    """Test that shape as int works."""
    strategy = ust.quantities(unit="m", shape=5)
    assert strategy is not None


def test_accepts_shape_tuple() -> None:
    """Test that shape as tuple works."""
    strategy = ust.quantities(unit="m", shape=(2, 3))
    assert strategy is not None


@given(q=ust.quantities(unit="eV", shape=()))
@settings(max_examples=20)
def test_explicit_scalar_shape(q: u.Quantity) -> None:
    """Test explicit scalar shape ()."""
    assert isinstance(q, u.Quantity)
    assert q.shape == ()
    assert u.dimension_of(q) == u.dimension("energy")


# Tests combining quantities and units strategies


@given(q=ust.quantities(unit=ust.units("length"), shape=3))
@settings(max_examples=50)
def test_length_varying_units(q: u.Quantity) -> None:
    """Test length quantity with varying units from units strategy."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    # Should always have length dimension
    assert u.dimension_of(q) == u.dimension("length")


@given(q=ust.quantities(unit=ust.units("velocity"), shape=()))
@settings(max_examples=50)
def test_scalar_velocity_varying_units(q: u.Quantity) -> None:
    """Test scalar velocity with varying units."""
    assert isinstance(q, u.Quantity)
    assert q.shape == ()
    assert u.dimension_of(q) == u.dimension("velocity")


@given(q=ust.quantities(unit=ust.units("mass"), shape=(2, 3)))
@settings(max_examples=30)
def test_2d_mass_varying_units(q: u.Quantity) -> None:
    """Test 2D mass quantity with varying units."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (2, 3)
    assert u.dimension_of(q) == u.dimension("mass")


@given(q=ust.quantities(unit=ust.units("energy", max_complexity=3), dtype=jnp.float32))
@settings(max_examples=30)
def test_energy_complex_units(q: u.Quantity) -> None:
    """Test energy with complex compound units."""
    assert isinstance(q, u.Quantity)
    assert q.dtype == jnp.float32
    assert u.dimension_of(q) == u.dimension("energy")


@given(
    q=ust.quantities(
        unit=ust.units("length", integer_powers=False),
        shape=5,
        dtype=jnp.float32,
    )
)
@settings(max_examples=30)
def test_length_non_integer_powers(q: u.Quantity) -> None:
    """Test length with non-integer power units."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (5,)
    assert q.dtype == jnp.float32
    assert u.dimension_of(q) == u.dimension("length")


# ==============================================================================
# Tests for dtype as a strategy
# ==============================================================================


@given(q=ust.quantities("m", dtype=st.sampled_from([jnp.float32, jnp.int32])))
@settings(max_examples=50)
def test_dtype_strategy_float_types(q: u.Quantity) -> None:
    """Test dtype as a strategy with different types."""
    assert isinstance(q, u.Quantity)
    assert q.dtype in (jnp.float32, jnp.int32)
    assert u.dimension_of(q) == u.dimension("length")


@given(
    q=ust.quantities(
        unit="kg",
        shape=(3,),
        dtype=st.sampled_from([jnp.float32, jnp.complex64]),
    )
)
@settings(max_examples=30)
def test_dtype_strategy_with_complex(q: u.Quantity) -> None:
    """Test dtype strategy including complex types."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    assert q.dtype in (jnp.float32, jnp.complex64)


@given(
    q=ust.quantities(
        unit="m/s",
        shape=(2, 3),
        dtype=st.just(jnp.float32),
    )
)
@settings(max_examples=20)
def test_dtype_strategy_just_float32(q: u.Quantity) -> None:
    """Test using st.just for a single dtype via strategy."""
    assert isinstance(q, u.Quantity)
    assert q.dtype == jnp.float32
    assert q.shape == (2, 3)


# ==============================================================================
# Tests for unit as a dimension
# ==============================================================================


@given(q=ust.quantities(unit=u.dimension("length"), shape=3))
@settings(max_examples=50)
def test_unit_from_length_dimension(q: u.Quantity) -> None:
    """Test generating units from length dimension."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    assert u.dimension_of(q) == u.dimension("length")


@given(q=ust.quantities(unit=u.dimension("velocity"), shape=()))
@settings(max_examples=50)
def test_unit_from_velocity_dimension(q: u.Quantity) -> None:
    """Test generating units from velocity dimension."""
    assert isinstance(q, u.Quantity)
    assert q.shape == ()
    assert u.dimension_of(q) == u.dimension("velocity")


@given(q=ust.quantities(unit=u.dimension("mass"), shape=(2, 3), dtype=jnp.float32))
@settings(max_examples=30)
def test_unit_from_mass_dimension_with_dtype(q: u.Quantity) -> None:
    """Test dimension with explicit dtype."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (2, 3)
    assert q.dtype == jnp.float32
    assert u.dimension_of(q) == u.dimension("mass")


@given(q=ust.quantities(unit=u.dimension("energy"), shape=5))
@settings(max_examples=30)
def test_unit_from_energy_dimension(q: u.Quantity) -> None:
    """Test generating units from energy dimension."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (5,)
    assert u.dimension_of(q) == u.dimension("energy")


# ==============================================================================
# Tests for unit as a dimension strategy
# ==============================================================================


@given(
    q=ust.quantities(
        st.sampled_from([u.dimension("length"), u.dimension("mass")]),
        shape=(),
    )
)
@settings(max_examples=50)
def test_unit_from_dimension_strategy(q: u.Quantity) -> None:
    """Test generating units from a strategy of dimensions."""
    assert isinstance(q, u.Quantity)
    assert q.shape == ()
    dim = u.dimension_of(q)
    assert dim in (u.dimension("length"), u.dimension("mass"))


@given(
    q=ust.quantities(
        st.sampled_from([u.dimension("velocity"), u.dimension("acceleration")]),
        shape=3,
        dtype=jnp.float32,
    )
)
@settings(max_examples=40)
def test_dimension_strategy_kinematic_quantities(q: u.Quantity) -> None:
    """Test dimension strategy for kinematic quantities."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3,)
    assert q.dtype == jnp.float32
    dim = u.dimension_of(q)
    assert dim in (u.dimension("velocity"), u.dimension("acceleration"))


@given(
    q=ust.quantities(
        st.sampled_from([u.dimension("angle"), u.dimension("dimensionless")]),
        shape=(2, 2),
    )
)
@settings(max_examples=30)
def test_dimension_strategy_angle_dimensionless(q: u.Quantity) -> None:
    """Test dimension strategy with angle and dimensionless."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (2, 2)
    dim = u.dimension_of(q)
    assert dim in (u.dimension("angle"), u.dimension("dimensionless"))


# ==============================================================================
# Tests combining dtype and dimension strategies
# ==============================================================================


@given(
    q=ust.quantities(
        unit=st.sampled_from([u.dimension("length"), u.dimension("time")]),
        dtype=st.sampled_from([jnp.float32, jnp.int32]),
        shape=(),
    )
)
@settings(max_examples=40)
def test_combined_dimension_and_dtype_strategies(q: u.Quantity) -> None:
    """Test combining dimension and dtype strategies."""
    assert isinstance(q, u.Quantity)
    assert q.shape == ()
    assert q.dtype in (jnp.float32, jnp.int32)
    dim = u.dimension_of(q)
    assert dim in (u.dimension("length"), u.dimension("time"))


@given(
    q=ust.quantities(
        unit=u.dimension("force"),
        dtype=st.sampled_from([jnp.float32, jnp.complex64]),
        shape=(3, 3),
    )
)
@settings(max_examples=30)
def test_force_dimension_with_dtype_strategy(q: u.Quantity) -> None:
    """Test force dimension with varying dtype."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (3, 3)
    assert q.dtype in (jnp.float32, jnp.complex64)
    assert u.dimension_of(q) == u.dimension("force")


# ==============================================================================
# Tests for elements parameter (value ranges)
# ==============================================================================


@given(
    q=ust.quantities(
        "kpc", shape=3, elements=st.floats(min_value=0, max_value=100, width=32)
    )
)
@settings(max_examples=50)
def test_positive_distance_elements(q: u.Quantity) -> None:
    """Test elements parameter for positive distances."""
    assert isinstance(q, u.Quantity)
    assert q.unit == u.unit("kpc")
    assert q.shape == (3,)
    # All values should be non-negative (suitable for distances)
    assert jnp.all(q.value >= 0)
    assert jnp.all(q.value <= 100)


@given(
    q=ust.quantities(
        "deg",
        shape=(),
        elements=st.floats(min_value=0, max_value=360, allow_nan=False, width=32),
    )
)
@settings(max_examples=50)
def test_longitude_angle_range(q: u.Quantity) -> None:
    """Test elements for longitude angles (0-360 degrees)."""
    assert isinstance(q, u.Quantity)
    assert q.unit == u.unit("deg")
    assert 0 <= q.value <= 360


@given(
    q=ust.quantities(
        "deg",
        shape=100,
        elements=st.floats(min_value=-90, max_value=90, allow_nan=False, width=32),
    )
)
@settings(max_examples=30)
def test_latitude_angle_range(q: u.Quantity) -> None:
    """Test elements for latitude angles (-90 to 90 degrees)."""
    assert isinstance(q, u.Quantity)
    assert q.unit == u.unit("deg")
    assert q.shape == (100,)
    assert jnp.all(q.value >= -90)
    assert jnp.all(q.value <= 90)


@given(
    q=ust.quantities(
        "rad",
        shape=(5, 5),
        elements=st.floats(min_value=0, max_value=6.0, allow_nan=False, width=32),
    )
)
@settings(max_examples=30)
def test_azimuthal_angle_radians(q: u.Quantity) -> None:
    """Test elements for azimuthal angles in radians (0 to 2π)."""
    assert isinstance(q, u.Quantity)
    assert q.unit == u.unit("rad")
    assert q.shape == (5, 5)
    assert jnp.all(q.value >= 0)
    assert jnp.all(q.value <= 6.0)


@given(
    q=ust.quantities(
        "m",
        shape=10,
        elements=st.floats(min_value=1.0, max_value=1000.0, allow_nan=False, width=32),
    )
)
@settings(max_examples=30)
def test_physical_length_scale(q: u.Quantity) -> None:
    """Test elements for physical lengths with realistic bounds."""
    assert isinstance(q, u.Quantity)
    assert q.shape == (10,)
    # All values in reasonable range (1 meter to 1 kilometer)
    assert jnp.all(q.value >= 1.0)
    assert jnp.all(q.value <= 1000.0)


# ==============================================================================
# Tests for wrap_to strategy
# ==============================================================================


@given(
    angle=ust.wrap_to(
        ust.quantities("deg", quantity_cls=u.Angle),
        min=u.Q(0, "deg"),
        max=u.Q(360, "deg"),
    )
)
@settings(max_examples=50, deadline=None)
def test_wrap_to_longitude_range(angle: u.Angle) -> None:
    """Test wrapping angles to longitude range [0, 360)."""
    assert isinstance(angle, u.Angle)
    # wrap_to wraps to [min, max), but floating point can give us max
    # Use ustrip to properly handle unit conversions
    angle_deg = u.ustrip("deg", angle)
    assert 0 <= angle_deg <= 360


@given(
    angle=ust.wrap_to(
        ust.quantities("rad", quantity_cls=u.Angle, shape=10),
        min=u.Q(0, "rad"),
        max=u.Q(6.28, "rad"),
    )
)
@settings(max_examples=30)
def test_wrap_to_angle_array(angle: u.Angle) -> None:
    """Test wrapping angle arrays to range."""
    assert isinstance(angle, u.Angle)
    assert angle.shape == (10,)
    # wrap_to wraps to [min, max), but floating point can give us max
    # Use ustrip to properly handle unit conversions
    angle_rad = u.ustrip("rad", angle)
    assert jnp.all(angle_rad >= 0)
    assert jnp.all(angle_rad <= 6.28)


@given(
    lat=ust.wrap_to(
        ust.quantities("deg", quantity_cls=u.Angle, shape=()),
        min=u.Q(-90, "deg"),
        max=u.Q(90, "deg"),
    )
)
@settings(max_examples=30)
def test_wrap_to_latitude_range(lat: u.Angle) -> None:
    """Test wrapping angles to latitude range [-90, 90)."""
    assert isinstance(lat, u.Angle)
    # wrap_to wraps to [min, max), but floating point can give us max
    # Use ustrip to properly handle unit conversions
    lat_deg = u.ustrip("deg", lat)
    assert -90 <= lat_deg <= 90


# Tests for quantity_cls parameter
# ==============================================================================


@given(angle=ust.quantities("rad", quantity_cls=u.Angle, shape=3))
@settings(max_examples=30)
def test_quantities_angle_class(angle: u.Angle) -> None:
    """Test creating Angle instances via quantity_cls parameter."""
    assert isinstance(angle, u.Angle)
    assert angle.unit == u.unit("rad")
    assert angle.shape == (3,)


@given(quantity=ust.quantities("m", quantity_cls=u.Quantity, shape=()))
@settings(max_examples=30)
def test_quantities_quantity_class(quantity: u.Quantity) -> None:
    """Test creating Quantity instances via quantity_cls parameter."""
    assert isinstance(quantity, u.Quantity)
    assert quantity.unit == u.unit("m")
    assert quantity.shape == ()



"""Tests for the units strategy."""
# pylint: disable=unreachable

import astropy.units as apyu
import hypothesis.strategies as st
from hypothesis import given, settings

import unxt as u
import unxt_hypothesis as ust


@given(unit=ust.units("length"))
@settings(max_examples=50)
def test_length_dimension(unit: u.AbstractUnit) -> None:
    """Test that generated length units have the correct dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("length"), f"Expected length, got {dim}"


@given(unit=ust.units("velocity"))
@settings(max_examples=50)
def test_velocity_dimension(unit: u.AbstractUnit) -> None:
    """Test that generated velocity units have the correct dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("velocity"), f"Expected velocity, got {dim}"


@given(unit=ust.units("energy"))
@settings(max_examples=50)
def test_energy_dimension(unit: u.AbstractUnit) -> None:
    """Test that generated energy units have the correct dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("energy"), f"Expected energy, got {dim}"


@given(unit=ust.units("mass"))
@settings(max_examples=50)
def test_mass_dimension(unit: u.AbstractUnit) -> None:
    """Test that generated mass units have the correct dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("mass"), f"Expected mass, got {dim}"


@given(unit=ust.units("time"))
@settings(max_examples=50)
def test_time_dimension(unit: u.AbstractUnit) -> None:
    """Test that generated time units have the correct dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("time"), f"Expected time, got {dim}"


@given(unit=ust.units("force"))
@settings(max_examples=50)
def test_force_dimension(unit: u.AbstractUnit) -> None:
    """Test that generated force units have the correct dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("force"), f"Expected force, got {dim}"


@given(unit=ust.units("length", integer_powers=True, max_complexity=0))
@settings(max_examples=30)
def test_no_complexity(unit: u.AbstractUnit) -> None:
    """Test that max_complexity=0 only generates base and composed units."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("length")
    # Should only be base unit or composed forms, no compound cancellations


@given(unit=ust.units("velocity", max_complexity=5))
@settings(max_examples=30)
def test_high_complexity(unit: u.AbstractUnit) -> None:
    """Test that max_complexity allows compound units."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("velocity")


@given(unit=ust.units("energy", integer_powers=False))
@settings(max_examples=30)
def test_non_integer_powers(unit: u.AbstractUnit) -> None:
    """Test that integer_powers=False works."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("energy")


def test_accepts_string_dimension() -> None:
    """Test that the strategy accepts string dimensions."""
    strategy = ust.units("length")
    # Should not raise an error
    assert strategy is not None


def test_accepts_physical_type() -> None:
    """Test that the strategy accepts PhysicalType dimensions."""
    dim = u.dimension("length")
    strategy = ust.units(dim)
    # Should not raise an error
    assert strategy is not None


@given(unit=ust.units("acceleration"))
@settings(max_examples=30)
def test_compound_dimension(unit: u.AbstractUnit) -> None:
    """Test with compound dimension (acceleration = length/time^2)."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("acceleration")


@given(unit=ust.units("angle"))
@settings(max_examples=30)
def test_dimensionless_angle(unit: u.AbstractUnit) -> None:
    """Test with angular dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("angle")


@given(unit=ust.units("pressure"))
@settings(max_examples=30)
def test_pressure_dimension(unit: u.AbstractUnit) -> None:
    """Test with pressure dimension."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("pressure")


@given(unit=ust.units(st.sampled_from(["length", "time", "mass"])))
@settings(max_examples=50)
def test_sampled_from_dimension(unit: u.AbstractUnit) -> None:
    """Test that dimension can be a st.sampled_from strategy."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    expected_dims = [u.dimension("length"), u.dimension("time"), u.dimension("mass")]
    assert dim in expected_dims, f"Expected one of {expected_dims}, got {dim}"


# Tests for derived_units


@given(unit=ust.derived_units("m"))
@settings(max_examples=50)
def test_derived_units_from_string(unit: u.AbstractUnit) -> None:
    """Test that derived_units works with string base."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("length")


@given(unit=ust.derived_units(apyu.Unit("kg")))
@settings(max_examples=50)
def test_derived_units_from_unit(unit: u.AbstractUnit) -> None:
    """Test that derived_units works with Unit base."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("mass")


@given(unit=ust.derived_units(ust.units("velocity")))
@settings(max_examples=50)
def test_derived_units_from_strategy(unit: u.AbstractUnit) -> None:
    """Test that derived_units works with strategy base."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("velocity")


@given(unit=ust.derived_units("m", max_complexity=0))
@settings(max_examples=30)
def test_derived_units_no_complexity(unit: u.AbstractUnit) -> None:
    """Test that max_complexity=0 limits to base and composed forms."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("length")


@given(unit=ust.derived_units("s", max_complexity=5))
@settings(max_examples=30)
def test_derived_units_high_complexity(unit: u.AbstractUnit) -> None:
    """Test that max_complexity allows more compound units."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("time")


@given(unit=ust.derived_units("J", integer_powers=False))
@settings(max_examples=30)
def test_derived_units_non_integer_powers(unit: u.AbstractUnit) -> None:
    """Test that integer_powers=False allows fractional powers."""
    assert isinstance(unit, u.AbstractUnit)
    dim = u.dimension_of(unit)
    assert dim == u.dimension("energy")



"""Tests for named dimension strategies."""

import hypothesis.strategies as st
from hypothesis import given, settings

import unxt as u
import unxt_hypothesis as ust


def test_DIMENSION_NAMES_public() -> None:  # noqa: N802
    """Ensure the named dimension list is public and unique."""
    names = ust.DIMENSION_NAMES
    assert isinstance(names, tuple)
    assert names
    assert len(names) == len(set(names))


@given(name=st.sampled_from(ust.DIMENSION_NAMES))
@settings(max_examples=30)
def test_all_names_resolve(name: str) -> None:
    """All named dimensions should resolve via ``unxt.dimension``."""
    dim = u.dimension(name)
    assert isinstance(dim, u.AbstractDimension)


@given(dim=ust.named_dimensions())
@settings(max_examples=30)
def test_named_dimensions_strategy(dim: u.AbstractDimension) -> None:
    """Strategy should return valid dimensions."""
    assert isinstance(dim, u.AbstractDimension)



"""Utilities.

Copyright (c) 2023 Galactic Dynamics. All rights reserved.
"""

__all__: tuple[str, ...] = ()

from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast, runtime_checkable

from jax import dtypes
from jax.numpy import dtype as DType  # noqa: N812

import quaxed.lax as qlax

if TYPE_CHECKING:
    from typing import Self


_singleton_insts: dict[type, object] = {}


class SingletonMixin:
    """Singleton class.

    This class is a mixin that can be used to create singletons.

    Examples
    --------
    >>> class MySingleton(SingletonMixin):
    ...     pass

    >>> a = MySingleton()
    >>> b = MySingleton()
    >>> a is b
    True

    """

    def __new__(cls, /, *_: Any, **__: Any) -> "Self":
        # Check if instance already exists
        if cls in _singleton_insts:
            return cast("Self", _singleton_insts[cls])
        # Create new instance and cache it
        self = object.__new__(cls)
        _singleton_insts[cls] = self
        return self


@runtime_checkable
class HasDType(Protocol):
    """Protocol for objects that have a dtype attribute."""

    @property
    def dtype(self) -> DType:
        """The dtype of the object."""


HasDTypeT = TypeVar("HasDTypeT", bound=HasDType)


def promote_dtypes(*arrays: HasDTypeT) -> tuple[HasDTypeT, ...]:
    """Promotes all input arrays to a common dtype.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u

    >>> x1 = jnp.array([1, 2, 3], dtype=jnp.int32)
    >>> x2 = jnp.array([4, 5, 6], dtype=jnp.float32)

    >>> x1, x2 = promote_dtypes(x1, x2)
    >>> x1.dtype, x2.dtype
    (dtype('float32'), dtype('float32'))

    >>> q1 = u.Quantity.from_([1, 2, 3], "m", dtype=int)
    >>> q2 = u.Q([4.0, 5, 6], unit="km")
    >>> q1, q2 = promote_dtypes(q1, q2)
    >>> q1.dtype, q2.dtype
    (dtype('float32'), dtype('float32'))

    """
    common_dtype = dtypes.result_type(*arrays)
    # TODO: check if this copies.
    return tuple([qlax.convert_element_type(arr, common_dtype) for arr in arrays])  # type: ignore[arg-type,misc]  # pylint: disable=R1728


def promote_dtypes_if_needed(
    original_dtypes: tuple[DType, ...], /, *args: HasDTypeT
) -> tuple[HasDTypeT, ...]:
    """Promotes dtypes of args if they differ from original_dtypes.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> from unxt._src.utils import promote_dtypes_if_needed
    >>> x1 = jnp.array([1, 2, 3], dtype=jnp.int32)
    >>> x2 = jnp.array([4, 5, 6], dtype=jnp.float32)
    >>> original_dtypes = (x1.dtype, x1.dtype)
    >>> x1_new, x2_new = promote_dtypes_if_needed(original_dtypes, x1, x2)
    >>> x1_new.dtype, x2_new.dtype
    (dtype('float32'), dtype('float32'))

    """
    # Compare equality of all `original_dtypes` and all `args` dtypes.
    all_orig_dtypes_eq = all(dt == original_dtypes[0] for dt in original_dtypes)
    all_new_dtypes_eq = all(arg.dtype == args[0].dtype for arg in args)
    if all_orig_dtypes_eq and not all_new_dtypes_eq:
        return promote_dtypes(*args)

    return args



"""Test `quaxed.lax` functions."""
# ruff:noqa: N816

import operator as ops

import jax.numpy as jnp
import jax.tree as jtu
import pytest
from astropy.units import UnitBase
from jax import lax

import quaxed.lax as qlax

import unxt as u
from unxt.quantity import is_any_quantity

# ------------
# Pytest marks

mark_todo = pytest.mark.skip(reason="TODO: Implement this function.")

# ------------

x_val = jnp.array([[1, 2], [3, 4]], dtype=float)
x_L = u.Q(x_val, unit="m")
x_ND = u.Q(x_val, unit="")

y_val = jnp.array([[5, 6], [7, 8]], dtype=float)
y_L = u.Q(y_val, unit="m")
y_ND = u.Q(y_val, unit="")

xtrig_val = x_val / 10
xtrig = u.Q(xtrig_val, unit="")

xbit_val = jnp.array([[1, 0], [1, 0]], dtype=int)
xbit = u.Q(xbit_val, unit="")

xcomplex_val = jnp.array([[1 + 2j, 3 + 4j], [5 + 6j, 7 + 8j]], dtype=complex)
xcomplex = u.Q(xcomplex_val, unit="m")

conv_kernel = u.Q(jnp.array([[[[1.0, 0.0], [0.0, -1.0]]]], dtype=float), unit="m")

xround_val = jnp.array([[1.1, 2.2], [3.3, 4.4]], dtype=float)
xround = u.Q(xround_val, unit="m")


# ------------


@pytest.mark.parametrize(
    ("func_name", "args", "kw", "expected"),
    [
        ("abs", (x_L,), {}, u.Q(lax.abs(x_val), unit="m")),
        ("acos", (xtrig,), {}, u.Q(lax.acos(xtrig_val), unit="rad")),
        ("acosh", (x_ND,), {}, u.Q(lax.acosh(x_val), unit="rad")),
        ("add", (x_L, y_L), {}, u.Q(lax.add(x_val, y_val), unit="m")),
        pytest.param("after_all", (), {}, True, marks=mark_todo),
        (
            "approx_max_k",
            (x_L,),
            {"k": 2},
            [u.Q([[2.0, 1], [4, 3]], unit="m"), u.Q([[1.0, 0], [1, 0]], unit="m")],
        ),
        (
            "approx_min_k",
            (x_L,),
            {"k": 2},
            [
                u.Q([[1.0, 2.0], [3.0, 4.0]], unit="m"),
                u.Q([[0.0, 1.0], [0.0, 1.0]], unit="m"),
            ],
        ),
        ("argmax", (x_L,), {"axis": 0, "index_dtype": int}, jnp.array([1, 1])),
        ("argmin", (x_L,), {"axis": 0, "index_dtype": int}, jnp.array([0, 0])),
        ("asin", (xtrig,), {}, u.Q(lax.asin(xtrig_val), unit="rad")),
        ("asinh", (xtrig,), {}, u.Q(lax.asinh(xtrig_val), unit="rad")),
        ("atan", (xtrig,), {}, u.Q(lax.atan(xtrig_val), unit="rad")),
        ("atan2", (x_L, y_L), {}, u.Q(lax.atan2(x_val, y_val), unit="rad")),
        ("atanh", (xtrig,), {}, u.Q(lax.atanh(xtrig_val), unit="rad")),
        (
            "batch_matmul",
            (
                u.Q(
                    jnp.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]], dtype=float),
                    unit="m",
                ),
                u.Q(
                    jnp.array([[[9, 10], [11, 12]], [[13, 14], [15, 16]]], dtype=float),
                    unit="m",
                ),
            ),
            {},
            u.Q(
                [[[31.0, 34.0], [71.0, 78.0]], [[155.0, 166.0], [211.0, 226.0]]],
                unit="m2",
            ),
        ),
        ("bessel_i0e", (x_ND,), {}, u.Q(lax.bessel_i0e(x_val), unit="")),
        ("bessel_i1e", (x_ND,), {}, u.Q(lax.bessel_i1e(x_val), unit="")),
        (
            "betainc",
            (1.0, xtrig, xtrig),
            {},
            u.Q(lax.betainc(1.0, xtrig_val, xtrig_val), ""),
        ),
        (
            "bitcast_convert_type",
            (x_L, jnp.float16),
            {},
            u.Q(lax.bitcast_convert_type(x_val, jnp.float16), unit="m"),
        ),
        ("bitwise_and", (xbit, xbit), {}, lax.bitwise_and(xbit_val, xbit_val)),
        ("bitwise_not", (xbit,), {}, u.Q(lax.bitwise_not(xbit_val), unit="")),
        ("bitwise_or", (xbit, xbit), {}, u.Q(lax.bitwise_or(xbit_val, xbit_val), "")),
        ("bitwise_xor", (xbit, xbit), {}, u.Q(lax.bitwise_xor(xbit_val, xbit_val), "")),
        (
            "broadcast",
            (x_L,),
            {"sizes": (2, 2)},
            u.Q(lax.broadcast(x_val, (2, 2)), unit="m"),
        ),
        (
            "broadcast_in_dim",
            (x_L, (1, 1, 2, 2), (2, 3)),
            {},
            u.Q(lax.broadcast_in_dim(x_val, (1, 1, 2, 2), (2, 3)), unit="m"),
        ),
        (
            "broadcast_to_rank",
            (x_L,),
            {"rank": 3},
            u.Q(lax.broadcast_to_rank(x_val, rank=3), unit="m"),
        ),
        pytest.param("broadcasted_iota", (), {}, True, marks=mark_todo),
        ("cbrt", (x_L,), {}, u.Q(lax.cbrt(x_val), unit=u.unit("m**(1/3)"))),
        ("ceil", (x_L,), {}, u.Q(lax.ceil(x_val), unit="m")),
        (
            "clamp",
            (u.Q(2.0, "m"), x_L, u.Q(3.0, "m")),
            {},
            u.Q(lax.clamp(2.0, x_val, 3.0), unit="m"),
        ),
        (
            "clz",
            (x_L.astype(int),),
            {},
            u.Q(lax.clz(x_val.astype(int)), unit="m"),
        ),
        ("collapse", (x_L, 1), {}, u.Q(lax.collapse(x_val, 1), unit="m")),
        (
            "concatenate",
            ((x_L, y_L), 0),
            {},
            u.Q(jnp.concatenate((x_val, y_val), 0), unit="m"),
        ),
        ("conj", (xcomplex,), {}, u.Q(lax.conj(xcomplex_val), unit="m")),
        pytest.param(
            "conv",
            (
                u.Q(jnp.arange(1, 17, dtype=float).reshape((1, 1, 4, 4)), unit="m"),
                conv_kernel,
            ),
            {"window_strides": (1, 1), "padding": "SAME"},
            None,
            marks=mark_todo,
        ),
        (
            "convert_element_type",
            (x_L, jnp.int16),
            {},
            u.Q(lax.convert_element_type(x_val, jnp.int16), unit="m"),
        ),
        pytest.param(
            "conv_general_dilated",
            (
                u.Q(jnp.arange(1, 17, dtype=float).reshape((1, 1, 4, 4)), unit="m"),
                conv_kernel,
            ),
            {"window_strides": (1, 1), "padding": "SAME"},
            None,
            marks=mark_todo,
        ),
        pytest.param("conv_general_dilated_local", (), {}, True, marks=mark_todo),
        pytest.param(
            "conv_general_dilated_patches",
            (u.Q(jnp.arange(1, 17, dtype=float).reshape((1, 1, 4, 4)), unit="m"),),
            {"filter_shape": (2, 2), "window_strides": (1, 1), "padding": "VALID"},
            True,
            marks=mark_todo,
        ),
        pytest.param(
            "conv_transpose",
            (
                u.Q(jnp.arange(1, 17, dtype=float).reshape((1, 1, 4, 4)), unit="m"),
                conv_kernel,
            ),
            {
                "strides": (2, 2),
                "padding": "SAME",
                "dimension_numbers": ("NCHW", "OIHW", "NCHW"),
            },
            True,
            marks=mark_todo,
        ),
        pytest.param("conv_with_general_padding", (), {}, True, marks=mark_todo),
        ("cos", (x_ND,), {}, u.Q(lax.cos(x_val), unit="")),
        ("cosh", (x_ND,), {}, u.Q(lax.cosh(x_val), unit="")),
        ("cumlogsumexp", (x_L,), {"axis": 0}, u.Q(lax.cumlogsumexp(x_val), unit="m")),
        ("cummax", (x_L,), {"axis": 0}, u.Q(lax.cummax(x_val), unit="m")),
        ("cummin", (x_L,), {"axis": 0}, u.Q(lax.cummin(x_val), unit="m")),
        ("cumprod", (x_ND,), {"axis": 0}, u.Q(lax.cumprod(x_val), unit="")),
        ("cumsum", (x_L,), {"axis": 0}, u.Q(lax.cumsum(x_val), unit="m")),
        ("digamma", (x_ND,), {}, u.Q(lax.digamma(x_val), unit="")),
        ("div", (x_L, y_L), {}, u.Q(lax.div(x_val, y_val), unit="")),
        ("dot", (x_L, y_L), {}, u.Q(lax.dot(x_val, y_val), unit="m2")),
        pytest.param("dot_general", (), {}, True, marks=mark_todo),
        pytest.param("dynamic_index_in_dim", (), {}, True, marks=mark_todo),
        (
            "dynamic_slice",
            (x_L, (0, 0), (2, 2)),
            {},
            u.Q(lax.dynamic_slice(x_val, (0, 0), (2, 2)), unit="m"),
        ),
        pytest.param("dynamic_slice_in_dim", (), {}, True, marks=mark_todo),
        pytest.param("dynamic_update_index_in_dim", (), {}, True, marks=mark_todo),
        (
            "dynamic_update_slice",
            (x_L, y_L, (0, 0)),
            {},
            u.Q(lax.dynamic_update_slice(x_val, y_val, (0, 0)), unit="m"),
        ),
        (
            "dynamic_update_slice_in_dim",
            (x_L, y_L, 0, 0),
            {},
            u.Q(lax.dynamic_update_slice_in_dim(x_val, y_val, 0, 0), unit="m"),
        ),
        ("eq", (x_L, x_L), {}, lax.eq(x_val, x_val)),
        ("erf", (xtrig,), {}, u.Q(lax.erf(xtrig_val), unit="")),
        ("erfc", (xtrig,), {}, u.Q(lax.erfc(xtrig_val), unit="")),
        ("erf_inv", (xtrig,), {}, u.Q(lax.erf_inv(xtrig_val), unit="")),
        ("exp", (x_ND,), {}, u.Q(lax.exp(x_val), unit="")),
        ("exp2", (x_ND,), {}, u.Q(lax.exp2(x_val), unit="")),
        (
            "expand_dims",
            (x_L, (0,)),
            {},
            u.Q(lax.expand_dims(x_val, (0,)), unit="m"),
        ),
        ("expm1", (x_ND,), {}, u.Q(lax.expm1(x_val), unit="")),
        (
            "fft",
            (x_L,),
            {"fft_type": "fft", "fft_lengths": (2, 2)},
            u.Q(lax.fft(x_val, fft_type="fft", fft_lengths=(2, 2)), unit="m-1"),
        ),
        ("floor", (xround,), {}, u.Q(lax.floor(xround_val), "m")),
        ("full_like", (x_L, u.Q(1.0, "m")), {}, u.Q(lax.full_like(x_val, 1.0), "m")),
        pytest.param("gather", (), {}, True, marks=mark_todo),
        ("ge", (x_L, y_L), {}, lax.ge(x_val, y_val)),
        ("gt", (x_L, y_L), {}, lax.gt(x_val, y_val)),
        ("igamma", (1.0, xtrig), {}, u.Q(lax.igamma(1.0, xtrig_val), unit="")),
        ("igammac", (1.0, xtrig), {}, u.Q(lax.igammac(1.0, xtrig_val), unit="")),
        ("imag", (xcomplex,), {}, u.Q(lax.imag(xcomplex_val), unit="m")),
        ("index_in_dim", (x_L, 0, 0), {}, u.Q(lax.index_in_dim(x_val, 0, 0), unit="m")),
        pytest.param("index_take", (), {}, True, marks=mark_todo),
        ("integer_pow", (x_L, 2), {}, u.Q(lax.integer_pow(x_val, 2), unit="m2")),
        pytest.param("iota", (), {}, True, marks=mark_todo),
        ("is_finite", (x_L,), {}, lax.is_finite(x_val)),
        ("le", (x_L, y_L), {}, lax.le(x_val, y_val)),
        ("lgamma", (x_ND,), {}, u.Q(lax.lgamma(x_val), unit="")),
        ("log", (x_ND,), {}, u.Q(lax.log(x_val), unit="")),
        ("log1p", (x_ND,), {}, u.Q(lax.log1p(x_val), unit="")),
        ("logistic", (x_ND,), {}, u.Q(lax.logistic(x_val), unit="")),
        ("lt", (x_L, y_L), {}, lax.lt(x_val, y_val)),
        ("max", (x_L, y_L), {}, u.Q(lax.max(x_val, y_val), unit="m")),
        ("min", (x_L, y_L), {}, u.Q(lax.min(x_val, y_val), unit="m")),
        ("mul", (x_L, y_L), {}, u.Q(lax.mul(x_val, y_val), unit="m2")),
        ("ne", (x_L, y_L), {}, lax.ne(x_val, y_val)),
        ("neg", (x_L,), {}, u.Q(lax.neg(x_val), unit="m")),
        ("nextafter", (x_L, y_L), {}, u.Q(lax.nextafter(x_val, y_val), unit="m")),
        pytest.param("pad", (), {}, True, marks=mark_todo),
        ("polygamma", (1.0, xtrig), {}, u.Q(lax.polygamma(1.0, xtrig_val), unit="")),
        ("population_count", (xbit,), {}, u.Q(lax.population_count(xbit_val), unit="")),
        ("pow", (x_L, 2), {}, u.Q(lax.pow(x_val, 2), unit="m2")),
        pytest.param("random_gamma_grad", (1.0, x_ND), {}, True, marks=mark_todo),
        ("real", (xcomplex,), {}, u.Q(lax.real(xcomplex_val), unit="m")),
        ("reciprocal", (x_L,), {}, u.Q(lax.reciprocal(x_val), unit="m-1")),
        pytest.param("reduce", (), {}, True, marks=mark_todo),
        pytest.param("reduce_precision", (), {}, True, marks=mark_todo),
        pytest.param("reduce_window", (), {}, True, marks=mark_todo),
        ("rem", (x_L, y_L), {}, u.Q(lax.rem(x_val, y_val), unit="m")),
        (
            "reshape",
            (x_L, (1, 4)),
            {},
            u.Q(lax.reshape(x_val, (1, 4)), unit="m"),
        ),
        (
            "rev",
            (x_L,),
            {"dimensions": (0,)},
            u.Q(lax.rev(x_val, (0,)), unit="m"),
        ),
        pytest.param("rng_bit_generator", (), {}, True, marks=mark_todo),
        ("round", (xround,), {}, u.Q(lax.round(xround_val), unit="m")),
        ("rsqrt", (x_L**2,), {}, u.Q(lax.rsqrt(x_val**2), unit="m-1")),
        pytest.param("scatter", (), {}, True, marks=mark_todo),
        pytest.param("scatter_apply", (), {}, True, marks=mark_todo),
        pytest.param("scatter_max", (), {}, True, marks=mark_todo),
        pytest.param("scatter_min", (), {}, True, marks=mark_todo),
        pytest.param("scatter_mul", (), {}, True, marks=mark_todo),
        (
            "shift_left",
            (xbit, xbit),
            {},
            u.Q(lax.shift_left(xbit_val, xbit_val), unit=""),
        ),
        pytest.param("shift_right_arithmetic", (xbit, 1), {}, True, marks=mark_todo),
        pytest.param("shift_right_logical", (xbit, 1), {}, True, marks=mark_todo),
        ("sign", (x_L,), {}, lax.sign(x_val)),
        ("sin", (x_ND,), {}, u.Q(lax.sin(x_val), unit="")),
        ("sinh", (x_ND,), {}, u.Q(lax.sinh(x_val), unit="")),
        (
            "slice",
            (x_L, (0, 0), (2, 2)),
            {},
            u.Q(lax.slice(x_val, (0, 0), (2, 2)), unit="m"),
        ),
        (
            "slice_in_dim",
            (x_L, 0, 1, 2),
            {},
            u.Q(lax.slice_in_dim(x_val, 0, 1, 2), unit="m"),
        ),
        ("sort", (x_L,), {}, u.Q(lax.sort(x_val), "m")),
        pytest.param("sort_key_val", (), {}, True, marks=mark_todo),
        ("sqrt", (x_L**2,), {}, u.Q(lax.sqrt(x_val**2), "m")),
        ("square", (x_L,), {}, u.Q(lax.square(x_val), "m2")),
        ("sub", (x_L, y_L), {}, u.Q(lax.sub(x_val, y_val), "m")),
        ("tan", (x_ND,), {}, u.Q(lax.tan(x_val), "")),
        ("tanh", (x_ND,), {}, u.Q(lax.tanh(x_val), "")),
        ("top_k", (x_L, 1), {}, [u.Q([[2.0], [4.0]], "m"), u.Q([[1.0], [1.0]], "m")]),
        (
            "transpose",
            (x_L, (1, 0)),
            {},
            u.Q(lax.transpose(x_val, (1, 0)), unit="m"),
        ),
        ("zeta", (x_L, 2.0), {}, u.Q(lax.zeta(x_val, 2.0), unit="m")),
        pytest.param("associative_scan", (), {}, True, marks=mark_todo),
        pytest.param("fori_loop", (), {}, True, marks=mark_todo),
        pytest.param("scan", (), {}, True, marks=mark_todo),
        (
            "select",
            (jnp.array([[True, False], [True, False]], dtype=bool), x_L, y_L),
            {},
            u.Q(
                lax.select(
                    jnp.array([[True, False], [True, False]], dtype=bool), x_val, y_val
                ),
                unit="m",
            ),
        ),
        pytest.param("select_n", (), {}, True, marks=mark_todo),
        pytest.param("switch", (), {}, True, marks=mark_todo),
        (
            "while_loop",
            (lambda x: jnp.all(x < 10), lambda x: x + 1, x_ND),
            {},
            u.Q(
                lax.while_loop(lambda x: jnp.all(x < 10), lambda x: x + 1, x_val),
                unit="",
            ),
        ),
        ("stop_gradient", (x_L,), {}, u.Q(lax.stop_gradient(x_val), unit="m")),
        pytest.param("custom_linear_solve", (), {}, True, marks=mark_todo),
        pytest.param("custom_root", (), {}, True, marks=mark_todo),
        pytest.param("all_gather", (), {}, True, marks=mark_todo),
        pytest.param("all_to_all", (), {}, True, marks=mark_todo),
        pytest.param("psum", (), {}, True, marks=mark_todo),
        pytest.param("psum_scatter", (), {}, True, marks=mark_todo),
        pytest.param("pmax", (), {}, True, marks=mark_todo),
        pytest.param("pmin", (), {}, True, marks=mark_todo),
        pytest.param("pmean", (), {}, True, marks=mark_todo),
        pytest.param("ppermute", (), {}, True, marks=mark_todo),
        pytest.param("pshuffle", (), {}, True, marks=mark_todo),
        pytest.param("pswapaxes", (), {}, True, marks=mark_todo),
        pytest.param("axis_index", (), {}, True, marks=mark_todo),
        # # --- Sharding-related operators ---
        pytest.param("with_sharding_constraint", (), {}, True, marks=mark_todo),
    ],
)
def test_lax_functions(func_name, args, kw, expected):
    """Test lax vs qlax functions."""
    got = getattr(qlax, func_name)(*args, **kw)

    got_vals = jtu.map(
        lambda x: u.ustrip(x) if is_any_quantity(x) else x, got, is_leaf=is_any_quantity
    )
    got_units = jtu.map(u.unit_of, got, is_leaf=is_any_quantity)

    exp_vals = jtu.map(
        lambda x: u.ustrip(x) if is_any_quantity(x) else x,
        expected,
        is_leaf=is_any_quantity,
    )
    exp_units = jtu.map(u.unit_of, expected, is_leaf=is_any_quantity)

    assert jtu.all(jtu.map(jnp.allclose, got_vals, exp_vals))
    assert jtu.all(
        jtu.map(ops.eq, got_units, exp_units, is_leaf=lambda x: isinstance(x, UnitBase))
    )



"""Pytest configuration for unxt-api tests."""

from typing import Any

import pytest

# ==============================================================================
# Fixtures for custom types


@pytest.fixture
def custom_dimension_type():
    """Create a custom type with dimension support."""

    class CustomDimType:
        def __init__(self, dim_str: str) -> None:
            self.dim_str = dim_str

    return CustomDimType


@pytest.fixture
def custom_unit_type():
    """Create a custom type with unit support."""

    class CustomUnitType:
        def __init__(self, unit_str: str) -> None:
            self.unit_str = unit_str

    return CustomUnitType


@pytest.fixture
def custom_quantity_type():
    """Create a custom type with quantity support."""

    class CustomQuantity:
        def __init__(self, value: Any, unit_str: str) -> None:
            self.value = value
            self.unit_str = unit_str

    return CustomQuantity


@pytest.fixture
def custom_unitsystem_type():
    """Create a custom type with unit system support."""

    class CustomUnitSystem:
        def __init__(self, length: str, time: str) -> None:
            self.length_unit = length
            self.time_unit = time

    return CustomUnitSystem



"""Custom type annotations for astropy."""

__all__ = ("APYUnits",)


from typing import TypeAlias

import astropy.units as apyu

APYUnits: TypeAlias = (
    apyu.UnitBase | apyu.Unit | apyu.FunctionUnitBase | apyu.StructuredUnit
)



"""Advanced tests for unxt-api covering edge cases and advanced scenarios."""

import math

import plum
import pytest

import unxt_api as uapi

# ==============================================================================
# Dispatch method inspection tests
# ==============================================================================


def test_methods_attribute_exists() -> None:
    """Test that dispatch functions have methods attribute."""
    assert hasattr(uapi.dimension, "methods")
    assert hasattr(uapi.unit, "methods")
    assert hasattr(uapi.uconvert, "methods")


def test_can_list_registered_methods() -> None:
    """Test that we can list registered methods."""
    # All abstract functions should have at least their abstract signature
    assert len(uapi.dimension.methods) >= 0
    assert len(uapi.unit.methods) >= 0


def test_can_inspect_method_signatures() -> None:
    """Test that we can inspect individual method signatures."""
    # The methods should be inspectable
    for method in uapi.dimension.methods:
        assert hasattr(method, "signature")


# ==============================================================================
# Type annotation tests
# ==============================================================================


def test_abstract_signatures_use_any() -> None:
    """Test that abstract signatures accept Any type."""
    # Check dimension signature
    # Note: The actual signature may be wrapped, so we check the docstring
    assert uapi.dimension.__doc__ is not None


def test_return_type_annotations() -> None:
    """Test that functions have appropriate return type hints."""
    # is_unit_convertible should return bool
    # The others return Any in the abstract signatures
    # Type annotations are checked by mypy, not runtime


# ==============================================================================
# Concurrent dispatch registration tests
# ==============================================================================


def test_independent_registrations_dont_interfere() -> None:
    """Test that registrations for different types don't interfere."""

    class TypeA:
        value: str = "a"

    class TypeB:
        value: str = "b"

    # Register for TypeA
    @plum.dispatch
    def dimension(obj: TypeA, /) -> str:
        return "dimension_a"

    # Register for TypeB
    @plum.dispatch
    def dimension(obj: TypeB, /) -> str:
        return "dimension_b"

    # Both should work independently
    a = TypeA()
    b = TypeB()
    assert dimension(a) == "dimension_a"
    assert dimension(b) == "dimension_b"


def test_multiple_packages_can_register() -> None:
    """Test that multiple packages can register their types."""

    # Simulate package 1
    class Package1Type:
        pass

    @plum.dispatch
    def unit_of(obj: Package1Type, /) -> str:
        return "package1_unit"

    # Simulate package 2
    class Package2Type:
        pass

    @plum.dispatch
    def unit_of(obj: Package2Type, /) -> str:
        return "package2_unit"

    # Both should coexist
    obj1 = Package1Type()
    obj2 = Package2Type()
    assert unit_of(obj1) == "package1_unit"
    assert unit_of(obj2) == "package2_unit"


# ==============================================================================
# Complex type hierarchy tests
# ==============================================================================


def test_inheritance_hierarchy() -> None:
    """Test dispatch resolution with inheritance."""

    class BaseQuantity:
        pass

    class LengthQuantity(BaseQuantity):
        pass

    class DistanceQuantity(LengthQuantity):
        pass

    # Register for base
    @plum.dispatch
    def dimension_of(obj: BaseQuantity, /) -> str:
        return "base"

    # Register more specific
    @plum.dispatch
    def dimension_of(obj: LengthQuantity, /) -> str:
        return "length"

    # Test resolution
    base = BaseQuantity()
    length = LengthQuantity()
    distance = DistanceQuantity()

    assert dimension_of(base) == "base"
    assert dimension_of(length) == "length"
    # distance should resolve to length (most specific match)
    assert dimension_of(distance) == "length"


def test_multiple_inheritance() -> None:
    """Test dispatch with multiple inheritance."""

    class HasUnit:
        pass

    class HasDimension:
        pass

    class Quantity(HasUnit, HasDimension):
        pass

    # Register for each parent
    @plum.dispatch
    def unit_of(obj: HasUnit, /) -> str:
        return "has_unit"

    @plum.dispatch
    def dimension_of(obj: HasDimension, /) -> str:
        return "has_dimension"

    # Test
    q = Quantity()
    assert unit_of(q) == "has_unit"
    assert dimension_of(q) == "has_dimension"


# ==============================================================================
# Default behavior tests
# ==============================================================================


def test_is_unit_convertible_no_default() -> None:
    """Test that is_unit_convertible raises when no dispatch found."""

    class UnknownType:
        pass

    obj1 = UnknownType()
    obj2 = UnknownType()

    # Should raise when no dispatch is registered
    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.is_unit_convertible(obj1, obj2)


def test_other_functions_raise_by_default() -> None:
    """Test that other functions raise when no dispatch found."""

    class UnknownType:
        pass

    obj = UnknownType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.dimension(obj)

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.unit(obj)


# ==============================================================================
# Vararg function tests
# ==============================================================================


def test_ustrip_varargs() -> None:
    """Test that ustrip can handle variable arguments."""

    class SimpleQuantity:
        def __init__(self, value: float) -> None:
            self.value = value

    # Single argument dispatch
    @plum.dispatch
    def ustrip(q: SimpleQuantity, /) -> float:
        return q.value

    q = SimpleQuantity(42.0)
    assert ustrip(q) == 42.0


def test_wrap_to_keyword_redirect() -> None:
    """Test that wrap_to keyword argument version redirects correctly."""

    class TestAngle:
        def __init__(self, value: float) -> None:
            self.value = value

    # Positional version
    @plum.dispatch
    def wrap_to(x: TestAngle, min: TestAngle, max: TestAngle, /) -> TestAngle:
        return TestAngle(42.0)  # simplified

    # Test positional
    result = wrap_to(TestAngle(370), TestAngle(0), TestAngle(360))
    assert result.value == 42.0

    # Test keyword (should redirect to positional)
    result_kw = wrap_to(TestAngle(370), min=TestAngle(0), max=TestAngle(360))
    assert result_kw.value == 42.0


# ==============================================================================
# Edge case tests
# ==============================================================================


def test_none_values() -> None:
    """Test handling of None values."""

    class MaybeQuantity:
        def __init__(self, value: float | None) -> None:
            self.value = value

    @plum.dispatch
    def ustrip(q: MaybeQuantity, /) -> float | None:
        return q.value

    assert ustrip(MaybeQuantity(42.0)) == 42.0
    assert ustrip(MaybeQuantity(None)) is None


def test_zero_values() -> None:
    """Test handling of zero values."""

    class Quantity:
        def __init__(self, value: float) -> None:
            self.value = value

    @plum.dispatch
    def ustrip(q: Quantity, /) -> float:
        return q.value

    assert ustrip(Quantity(0.0)) == 0.0
    assert ustrip(Quantity(-0.0)) == -0.0


def test_special_float_values() -> None:
    """Test handling of special float values (inf, nan)."""

    class Quantity:
        def __init__(self, value: float) -> None:
            self.value = value

    @plum.dispatch
    def ustrip(q: Quantity, /) -> float:
        return q.value

    assert math.isinf(ustrip(Quantity(float("inf"))))
    assert math.isinf(ustrip(Quantity(float("-inf"))))
    assert math.isnan(ustrip(Quantity(float("nan"))))


# ==============================================================================
# Dispatch resolution order tests
# ==============================================================================


def test_most_specific_wins() -> None:
    """Test that the most specific dispatch is chosen."""

    class Base:
        pass

    class Derived(Base):
        pass

    @plum.dispatch
    def unit_of(obj: Base, /) -> str:
        return "base"

    @plum.dispatch
    def unit_of(obj: Derived, /) -> str:
        return "derived"

    assert unit_of(Base()) == "base"
    assert unit_of(Derived()) == "derived"


def test_ambiguous_dispatch_error() -> None:
    """Test that ambiguous dispatches raise AmbiguousLookupError."""

    # Plum raises an error for truly ambiguous cases
    class A:
        pass

    class B:
        pass

    class C(A, B):
        pass

    @plum.dispatch
    def dimension(obj: A, /) -> str:
        return "a"

    @plum.dispatch
    def dimension(obj: B, /) -> str:
        return "b"

    # C inherits from both A and B - this is ambiguous
    c = C()
    with pytest.raises(plum.resolver.AmbiguousLookupError):
        dimension(c)


# ==============================================================================
# Docstring example tests
# ==============================================================================


def test_uconvert_docstring_example_pattern() -> None:
    """Test that the pattern shown in uconvert docstring is valid."""
    # The docstring shows examples with unxt, but we test the pattern
    assert uapi.uconvert.__doc__ is not None
    assert "Examples" in uapi.uconvert.__doc__


def test_ustrip_docstring_example_pattern() -> None:
    """Test that the pattern shown in ustrip docstring is valid."""
    assert uapi.ustrip.__doc__ is not None
    assert "Examples" in uapi.ustrip.__doc__


def test_wrap_to_docstring_example_pattern() -> None:
    """Test that the pattern shown in wrap_to docstring is valid."""
    assert uapi.wrap_to.__doc__ is not None
    assert "Examples" in uapi.wrap_to.__doc__



"""Unit strategies."""

__all__ = ("derived_units", "units")

from typing import Any

import astropy.units as apyu
from hypothesis import strategies as st

import unxt as u
from .dimensions import named_dimensions


@st.composite
def derived_units(  # pylint: disable=unreachable
    draw: st.DrawFn,
    base: str | u.AbstractUnit | st.SearchStrategy[u.AbstractUnit],
    /,
    *,
    integer_powers: bool = True,
    max_complexity: int = 3,
) -> u.AbstractUnit:
    """Generate hypothesis strategy for units from a given base unit.

    This strategy generates units that are dimensionally equivalent to the input
    base unit by combining the base unit's base units and discovered composed
    forms.

    Parameters
    ----------
    draw : st.DrawFn
        Hypothesis draw function (automatically provided by @st.composite).
    base : str | u.AbstractUnit | st.SearchStrategy[u.AbstractUnit]
        Base unit (e.g., "m", "s", "kg") or a `hypothesis` strategy that
        generates such units.
    integer_powers : bool, optional
        If True, only generate units with integer powers of base units.  Default
        is True.
    max_complexity : int, optional
        Maximum number of additional base unit factors to combine.  Higher
        values create more complex compound units. Default is 3.

    Returns
    -------
    `unxt.AbstractUnit`
        A unit with the same physical dimension as the input.

    Examples
    --------
    >>> import hypothesis.strategies as st
    >>> from hypothesis import given
    >>> import unxt_hypothesis as ust
    >>> import unxt as u

    Generate units derived from a base unit:

    >>> @given(unit=ust.derived_units("m"))
    ... def test_length_units(unit):
    ...     assert u.dimension_of(unit) == u.dimension("length")

    Generate units from a strategy:

    >>> @given(unit=ust.derived_units(st.sampled_from(["km", "m/s^2"])))
    ... def test_velocity_derived(unit):
    ...     assert u.dimension_of(unit) in (
    ...         u.dimension("velocity"),
    ...         u.dimension("acceleration"),
    ...     )

    Control complexity of generated units:

    >>> @given(unit=ust.derived_units("kg", max_complexity=1))
    ... def test_simple_mass_units(unit):
    ...     assert u.dimension_of(unit) == u.dimension("mass")

    Allow non-integer powers:

    >>> @given(unit=ust.derived_units("m", integer_powers=False))
    ... def test_fractional_powers(unit):
    ...     dim = u.dimension_of(unit)
    ...     assert dim == u.dimension("length")

    """
    # Convert base to a unit if it's a string or draw from strategy
    base_unit = u.unit(draw(base) if isinstance(base, st.SearchStrategy) else base)

    # Collect all possible unit forms
    # 1. Start with the base unit
    candidates = [base_unit]

    # 2. Add composed forms (limited to first 20 to avoid huge search space)
    composed = list(base_unit.compose())[:20]
    candidates.extend(composed)

    # 3. Optionally create compound units by combining cancelling factors
    if max_complexity > 0:
        # Get SI base decomposition
        decomposed = base_unit.decompose()
        si_bases = list(decomposed.bases)
        list(decomposed.powers)

        # Get a pool of other SI base units to use as cancelling factors
        all_si_bases = [
            apyu.Unit(x) for x in ("m", "s", "kg", "A", "K", "mol", "cd", "rad")
        ]

        # Find SI bases not used in this dimension (for pure cancellation)
        # and those that are used (for modification)
        available_bases = [b for b in all_si_bases if b not in si_bases]

        if available_bases:
            # Generate a few compound units with cancelling factors
            num_extra = draw(st.integers(min_value=0, max_value=max_complexity))

            if num_extra > 0:
                # Choose random bases to combine
                extra_bases = draw(
                    st.lists(
                        st.sampled_from(available_bases),
                        min_size=num_extra,
                        max_size=num_extra,
                        unique=True,
                    )
                )

                # For each base, generate a power and its inverse
                compound = base_unit
                for extra_base in extra_bases:
                    if integer_powers:
                        power = draw(st.integers(min_value=1, max_value=3))
                    else:
                        power = draw(st.floats(min_value=0.5, max_value=3.0))

                    # Multiply and divide to cancel dimensions
                    compound = compound * (extra_base**power) / (extra_base**power)

                candidates.append(compound)

    # Choose one of the candidate units
    return draw(st.sampled_from(candidates))


@st.composite
def units(
    draw: st.DrawFn,
    dimension: str
    | u.AbstractDimension
    | st.SearchStrategy[u.AbstractDimension] = named_dimensions(),  # noqa: B008
    /,
    **kwargs: Any,
) -> u.AbstractUnit:
    """Generate hypothesis strategy for units with a given physical dimension.

    This strategy generates units that are dimensionally equivalent to the input
    dimension by combining the dimension's base units and discovered composed
    forms.

    Parameters
    ----------
    draw : st.DrawFn
        Hypothesis draw function (automatically provided by @st.composite).
    dimension : str | apyu.PhysicalType
        Physical dimension (e.g., "velocity", "length") or a `hypothesis`
        strategy that generates such dimensions. Default is a strategy sampling
        from SI base dimensions.
    **kwargs : Any
        Additional keyword arguments passed to `derived_units()`.
        For example:

        - integer_powers : bool
            If True, only generate units with integer powers of base units.  Default
            is True.
        - max_complexity : int
            Maximum number of additional base unit factors to combine.  Higher
            values create more complex compound units. Default is 3.

    Returns
    -------
    unxt.AbstractUnit
        A unit with the same physical dimension as the input.

    Examples
    --------
    >>> import hypothesis.strategies as st
    >>> from hypothesis import given
    >>> import unxt_hypothesis as ust
    >>> import unxt as u

    >>> @given(unit=ust.units("velocity"))
    ... def test_velocity_units(unit):
    ...     assert u.dimension_of(unit) == u.dimension("velocity")

    >>> @given(unit=ust.units("length", integer_powers=False))
    ... def test_length_units(unit):
    ...     dim = u.dimension_of(unit)
    ...     assert dim == u.dimension("length")

    The dimension argument can also be a strategy:

    >>> from hypothesis import strategies as st
    >>> @given(unit=ust.units(st.sampled_from(["length", "time", "mass"])))
    ... def test_multiple_dimensions(unit):
    ...     dim = u.dimension_of(unit)
    ...     assert dim in [
    ...         u.dimension("length"),
    ...         u.dimension("time"),
    ...         u.dimension("mass"),
    ...     ]

    Notes
    -----
    The strategy works by:

    1. Getting the canonical unit for the dimension
    2. Decomposing it into base SI units
    3. Finding composed forms using `.compose()`
    4. Generating random combinations by multiplying/dividing
       other units with cancelling dimensions

    """
    # Convert to Dimension object, drawing if necessary
    dimension = (
        u.dimension(draw(dimension))
        if isinstance(dimension, st.SearchStrategy)
        else u.dimension(dimension)
    )

    # Get the canonical unit for this dimension
    base_unit = dimension._unit  # noqa: SLF001

    return draw(derived_units(base_unit, **kwargs))



"""Tools for representing systems of units using ``astropy.units``."""

__all__ = ("UNITSYSTEMS_REGISTRY", "AbstractUnitSystem")

from collections.abc import Iterator
from dataclasses import dataclass
from types import MappingProxyType
from typing import ClassVar, get_args, get_type_hints

from astropy.units import PhysicalType, UnitBase as AstropyUnitBase
from astropy.units.physical import _physical_unit_mapping

from is_annotated import isannotated

from .utils import parse_dimlike_name
from unxt.dims import AbstractDimension, dimension
from unxt.units import AbstractUnit, unit

Unit = AstropyUnitBase

_UNITSYSTEMS_REGISTRY: dict[
    tuple[AbstractDimension, ...], type["AbstractUnitSystem"]
] = {}
UNITSYSTEMS_REGISTRY = MappingProxyType(_UNITSYSTEMS_REGISTRY)


@dataclass(frozen=True, slots=True, eq=True)
class AbstractUnitSystem:
    """Represents a system of units.

    This class behaves like a dictionary with keys set by physical types (i.e.
    "length", "velocity", "energy", etc.). If a unit for a particular physical
    type is not specified on creation, a composite unit will be created with the
    base units. See the examples below for some demonstrations.

    Examples
    --------
    If only base units are specified, any physical type specified as a key to
    this object will be composed out of the base units:

    >>> from unxt import unitsystem
    >>> usys = unitsystem("m", "s", "kg", "radian")
    >>> usys
    unitsystem(m, s, kg, rad)

    >>> usys["velocity"]
    Unit("m / s")

    This unit system defines energy:

    >>> usys = unitsystem("m", "s", "kg", "radian", "erg")
    >>> usys["energy"]
    Unit("erg")

    This is useful for Galactic dynamics where lengths and times are usually
    given in terms of ``kpc`` and ``Myr``, but velocities are often specified in
    ``km/s``:

    >>> usys = unitsystem("kpc", "Myr", "Msun", "radian", "km / s")
    >>> usys["velocity"]
    Unit("km / s")

    Unit systems can be hashed:

    >>> isinstance(hash(usys), int)
    True

    And iterated over:

    >>> [x for x in usys]
    [Unit("kpc"), Unit("Myr"), Unit("solMass"), Unit("rad"), Unit("km / s")]

    With length equal to the number of base units

    >>> len(usys)
    5

    """

    # ===============================================================
    # Class-level

    _base_field_names: ClassVar[tuple[str, ...]]
    _base_dimensions: ClassVar[tuple[AbstractDimension, ...]]

    def __init_subclass__(cls) -> None:
        # Register class with a tuple of it's dimensions.
        # This requires processing the type hints, not the dataclass fields
        # since those are made after the original class is defined.
        field_names, dims = parse_field_names_and_dimensions(cls)

        # Check the unitsystem is not already registered
        # If `make_dataclass(slots=True)` then the class is made twice, the
        # second time adding the `__slots__` attribute
        if dims in _UNITSYSTEMS_REGISTRY and "__slots__" not in cls.__dict__:
            msg = f"Unit system with dimensions {dims} already exists."
            raise ValueError(msg)

        # Add attributes to the class
        cls._base_field_names = tuple(field_names)
        cls._base_dimensions = dims

        _UNITSYSTEMS_REGISTRY[dims] = cls

    # ===============================================================
    # USys API

    @property  # TODO: classproperty
    def base_dimensions(self) -> tuple[AbstractDimension, ...]:
        """Dimensions required for the unit system."""
        return self._base_dimensions

    @property
    def base_units(self) -> tuple[AbstractUnit, ...]:
        """List of core units."""
        return tuple(getattr(self, k) for k in self._base_field_names)

    # ===============================================================
    # Python stuff

    def __getitem__(self, key: AbstractDimension | str) -> AbstractUnit:
        """Get the unit for a given physical type.

        Examples
        --------
        >>> from unxt import unitsystem
        >>> usys = unitsystem("m", "s", "kg", "radian")

        Something in the base dimensions:

        >>> usys["length"]
        Unit("m")

        Something not in the base dimensions:

        >>> usys["velocity"]
        Unit("m / s")

        """
        key = dimension(key)
        if key in self.base_dimensions:
            return getattr(self, parse_dimlike_name(key))

        out: AbstractUnit
        for k, v in _physical_unit_mapping.items():
            if v == key:
                out = unit(" ".join([f"{x}**{y}" for x, y in k]))
                break
        # IDK if it's possible to get here
        else:
            msg = f"Physical type {key!r} doesn't exist in unit registry."  # pragma: no cover  # noqa: E501
            raise ValueError(msg)  # pragma: no cover

        out = out.decompose(self.base_units)
        out._scale = 1.0  # noqa: SLF001
        return out

    def __len__(self) -> int:
        """Return the number of base units in the system.

        Examples
        --------
        >>> from unxt import unitsystem
        >>> usys = unitsystem("m", "s", "kg", "radian")
        >>> len(usys)
        4

        """
        # Note: This is required for q.decompose(usys) to work, where q is a Quantity
        return len(self.base_dimensions)

    # TODO: should this be changed to _base_field_names -> Iterator[str]?
    def __iter__(self) -> Iterator[AbstractUnit]:
        """Iterate over the base units.

        This is different than a dictionary, which would iterate over the
        physical types (the keys).

        Examples
        --------
        >>> from unxt import unitsystem
        >>> usys = unitsystem("m", "s", "kg", "radian")
        >>> list(iter(usys))
        [Unit("m"), Unit("s"), Unit("kg"), Unit("rad")]

        """
        yield from self.base_units

    def __str__(self) -> str:
        """Return a string representation of the unit system.

        Examples
        --------
        >>> from unxt import unitsystem
        >>> usys = unitsystem("m", "s", "kg", "radian")
        >>> str(usys)
        'LTMAUnitSystem(length, time, mass, angle)'

        """
        fs = ", ".join(map(str, self._base_field_names))
        return f"{type(self).__name__}({fs})"

    # ===============================================================
    # Plum stuff

    #: This tells `plum` that this type can be efficiently cached.
    __faithful__: ClassVar[bool] = True


# ---------------------------------------------------------------


def parse_field_names_and_dimensions(
    cls: type,
) -> tuple[tuple[str, ...], tuple[AbstractDimension, ...]]:
    # Register class with a tuple of it's dimensions.
    # This requires processing the type hints, not the dataclass fields
    # since those are made after the original class is defined.
    type_hints = get_type_hints(cls, include_extras=True)

    field_names = []
    dims = []
    for name, type_hint in type_hints.items():
        # Check it's Annotated
        if not isannotated(type_hint):
            continue

        # Get the arguments to Annotated
        origin, *f_args = get_args(type_hint)

        # Check that the first argument is a UnitBase
        if not issubclass(origin, Unit):
            continue

        # Need for one of the arguments to be a Dimension
        f_dim = [x for x in f_args if isinstance(x, PhysicalType)]
        if not f_dim:
            msg = f"Field {name!r} must be an Annotated with a dimension."
            raise TypeError(msg)
        if len(f_dim) > 1:
            msg = (
                f"Field {name!r} must be an Annotated with only one dimension; "
                f"got {f_dim}"
            )
            raise TypeError(msg)

        field_names.append(parse_dimlike_name(name))
        dims.append(f_dim[0])

    if len(set(dims)) < len(dims):
        msg = "Some dimensions are repeated."
        raise ValueError(msg)

    return tuple(field_names), tuple(dims)



"""Flags for unit system definitions."""

__all__ = ("AbstractUSysFlag", "DynamicalSimUSysFlag", "StandardUSysFlag")

from typing import Any


class AbstractUSysFlag:
    """Abstract class for unit system flags to provide dispatch control.

    Unit system flags are used to indicate the type of unit system being
    defined. They are not intended to be instantiated and are used to provide
    dispatch control for defining unit systems.

    Raises
    ------
    ValueError
        If an attempt is made to instantiate a unit system flag class.

    See Also
    --------
    `unxt.unitsystems.unitsystem` :
        Function to define a unit system. The `AbstractUSysFlag` can be provided
        as the first argument to indicate the type of unit system being defined.
        This is useful for multiple-dispatching to the correct constructor.

    Examples
    --------
    For this example we will use the `unxt.unitsystems.StandardUSysFlag`, which
    indicates the standard construction of a unit system directly from the set
    of base units. Flags like `unxt.unitsystems.DynamicalSimUSysFlag` can be
    used to create a unit system where the gravitational constant ``G`` is
    unit-less and units are defined accordingly.

    >>> import unxt

    Define a unit system with the standard flag:

    >>> unxt.unitsystem(unxt.unitsystems.StandardUSysFlag, "m", "kg", "s")
    LengthMassTimeUnitSystem(length=Unit("m"), mass=Unit("kg"), time=Unit("s"))

    """

    def __new__(cls, *_: Any, **__: Any) -> None:  # type: ignore[misc]
        msg = "unit system flag classes cannot be instantiated."
        raise ValueError(msg)


class StandardUSysFlag(AbstractUSysFlag):
    """Flag to indicate a standard unit system with no additional arguments.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem, StandardUSysFlag

    Define a unit system with the standard flag:

    >>> unitsystem(StandardUSysFlag, "m", "kg", "s")
    LengthMassTimeUnitSystem(length=Unit("m"), mass=Unit("kg"), time=Unit("s"))

    Further examples may be found in the ``unitsystem`` docs.

    """


class DynamicalSimUSysFlag(AbstractUSysFlag):
    """Flag to indicate a unit system with optional definition of G.

    Examples
    --------
    >>> from unxt.unitsystems import unitsystem, DynamicalSimUSysFlag

    Define a unit system with the dynamical simulation flag:

    >>> unitsystem(DynamicalSimUSysFlag, "m", "kg")
    LengthMassTimeUnitSystem(length=Unit("m"), mass=Unit("kg"), time=Unit("122404 s"))

    Further examples may be found in the ``unitsystem`` docs.

    """



"""Unit tests for dimension API (unxt_api._dimensions)."""

import plum
import pytest

import unxt_api as uapi

# ==============================================================================
# Tests for dimension()
# ==============================================================================


def test_dimension_is_abstract_dispatch() -> None:
    """Test that dimension is an abstract dispatch function."""
    assert isinstance(uapi.dimension, plum.function.Function)
    assert hasattr(uapi.dimension, "methods")


def test_dimension_accepts_any_type() -> None:
    """Test that dimension accepts Any type in signature."""
    # The abstract signature should accept Any
    assert "dimension" in dir(uapi)


def test_dimension_no_default_implementation() -> None:
    """Test that calling dimension without implementation raises error."""

    # Create a custom type that has no dispatch registered
    class NoDispatchType:
        pass

    obj = NoDispatchType()

    # Should raise NotFoundLookupError since no dispatch is registered
    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.dimension(obj)


def test_dimension_can_register_custom_dispatch(custom_dimension_type) -> None:
    """Test that custom dispatches can be registered."""

    # Register a custom dispatch
    @plum.dispatch
    def dimension(obj: custom_dimension_type, /) -> str:
        return obj.dim_str

    # Create instance and test
    obj = custom_dimension_type("length")
    result = dimension(obj)
    assert result == "length"


def test_dimension_multiple_dispatches_possible() -> None:
    """Test that multiple dispatches can coexist."""
    # The abstract function should support multiple dispatches
    # Count existing methods
    initial_count = len(uapi.dimension.methods)
    assert initial_count >= 0  # At least the abstract signature


# ==============================================================================
# Tests for dimension_of()
# ==============================================================================


def test_dimension_of_is_abstract_dispatch() -> None:
    """Test that dimension_of is an abstract dispatch function."""
    assert isinstance(uapi.dimension_of, plum.function.Function)
    assert hasattr(uapi.dimension_of, "methods")


def test_dimension_of_accepts_any_type() -> None:
    """Test that dimension_of accepts Any type in signature."""
    assert "dimension_of" in dir(uapi)


def test_dimension_of_no_default_implementation() -> None:
    """Test that calling dimension_of without implementation may return None.

    Note: When unxt is imported (e.g., during doctest execution), it may
    register a default implementation that returns None for unknown types.
    """

    class NoDispatchType:
        pass

    obj = NoDispatchType()

    try:
        result = uapi.dimension_of(obj)
        # If it doesn't raise, it should return None (unxt's default)
        assert result is None
    except plum.resolver.NotFoundLookupError:
        # This is also acceptable if unxt hasn't been imported
        pass


def test_dimension_of_can_register_custom_dispatch(custom_unit_type) -> None:
    """Test that custom dispatches can be registered for dimension_of."""

    @plum.dispatch
    def dimension_of(obj: custom_unit_type, /) -> str:
        # Simple implementation that extracts dimension from unit string
        return f"dim_of_{obj.unit_str}"

    obj = custom_unit_type("m")
    result = dimension_of(obj)
    assert result == "dim_of_m"


# ==============================================================================
# API consistency tests
# ==============================================================================


def test_both_dimension_functions_are_exported() -> None:
    """Test that both functions are exported from unxt_uapi."""
    assert hasattr(uapi, "dimension")
    assert hasattr(uapi, "dimension_of")


def test_both_dimension_functions_in_all() -> None:
    """Test that both functions are in __all__."""
    assert "dimension" in uapi.__all__
    assert "dimension_of" in uapi.__all__


def test_dimension_functions_have_independent_registries() -> None:
    """Test that dimension and dimension_of have independent dispatch registries."""
    # They should be different function objects
    assert uapi.dimension is not uapi.dimension_of

    # They should have their own method registries
    assert id(uapi.dimension.methods) != id(uapi.dimension_of.methods)



"""Usage tests demonstrating how to use unxt-api to create custom implementations.

These tests show realistic usage patterns for third-party packages that want
to integrate with the unxt dispatch system.
"""

from typing import Any, ClassVar

import plum
import pytest

import unxt_api as uapi

# ==============================================================================
# Custom type integration tests
# ==============================================================================


def test_simple_quantity_type() -> None:
    """Test creating a simple quantity type with unit support."""

    # Define a custom quantity type
    class SimpleQuantity:
        def __init__(self, value: float, unit: str) -> None:
            self.value = value
            self.unit = unit

        def __repr__(self) -> str:
            return f"SimpleQuantity({self.value}, '{self.unit}')"

    # Register dispatch for unit_of
    @plum.dispatch
    def unit_of(obj: SimpleQuantity, /) -> str:
        return obj.unit

    # Create instance and test
    q = SimpleQuantity(5.0, "m")
    assert unit_of(q) == "m"

    q2 = SimpleQuantity(10.0, "kg")
    assert unit_of(q2) == "kg"


def test_dimension_aware_type() -> None:
    """Test creating a type that supports both unit and dimension."""

    class DimensionalQuantity:
        _dimension_map: ClassVar = {
            "m": "length",
            "km": "length",
            "s": "time",
            "kg": "mass",
        }

        def __init__(self, value: float, unit: str) -> None:
            self.value = value
            self.unit = unit

        @property
        def dimension(self) -> str:
            return self._dimension_map.get(self.unit, "unknown")

    # Register dispatches
    @plum.dispatch
    def unit_of(obj: DimensionalQuantity, /) -> str:
        return obj.unit

    @plum.dispatch
    def dimension_of(obj: DimensionalQuantity, /) -> str:
        return obj.dimension

    # Test
    q = DimensionalQuantity(100.0, "km")
    assert unit_of(q) == "km"
    assert dimension_of(q) == "length"


def test_unit_conversion_implementation() -> None:
    """Test implementing unit conversion for a custom type."""

    class ConvertibleQuantity:
        _conversion_factors: ClassVar = {
            ("km", "m"): 1000.0,
            ("m", "km"): 0.001,
            ("m", "m"): 1.0,
            ("km", "km"): 1.0,
        }

        def __init__(self, value: float, unit: str) -> None:
            self.value = value
            self.unit = unit

        def convert_to(self, target_unit: str) -> "ConvertibleQuantity":
            factor = self._conversion_factors.get((self.unit, target_unit))
            if factor is None:
                msg = f"Cannot convert from {self.unit} to {target_unit}"
                raise ValueError(msg)
            return ConvertibleQuantity(self.value * factor, target_unit)

    # Register dispatch
    @plum.dispatch
    def uconvert(to_unit: str, quantity: ConvertibleQuantity, /) -> ConvertibleQuantity:
        return quantity.convert_to(to_unit)

    @plum.dispatch
    def is_unit_convertible(
        to_unit: str, from_quantity: ConvertibleQuantity, /
    ) -> bool:
        key = (from_quantity.unit, to_unit)
        return key in ConvertibleQuantity._conversion_factors

    # Test conversion
    q = ConvertibleQuantity(5.0, "km")
    result = uconvert("m", q)
    assert result.value == 5000.0
    assert result.unit == "m"

    # Test convertibility check
    assert is_unit_convertible("m", q) is True
    assert is_unit_convertible("kg", q) is False


def test_strip_units_implementation() -> None:
    """Test implementing unit stripping for a custom type."""

    class StrippableQuantity:
        def __init__(self, value: float, unit: str) -> None:
            self.value = value
            self.unit = unit

    # Register dispatch for basic ustrip
    @plum.dispatch
    def ustrip(quantity: StrippableQuantity, /) -> float:
        return quantity.value

    # Test
    q = StrippableQuantity(42.5, "m")
    assert ustrip(q) == 42.5


def test_angle_wrapping_implementation() -> None:
    """Test implementing angle wrapping for a custom type."""

    class Angle:
        def __init__(self, value: float, unit: str = "deg") -> None:
            self.value = value
            self.unit = unit

        def wrap(self, min_val: float, max_val: float) -> "Angle":
            range_size = max_val - min_val
            wrapped = ((self.value - min_val) % range_size) + min_val
            return Angle(wrapped, self.unit)

    # Register dispatch
    @plum.dispatch
    def wrap_to(x: Angle, min: Angle, max: Angle, /) -> Angle:
        return x.wrap(min.value, max.value)

    # Test
    angle = Angle(370.0, "deg")
    result = wrap_to(angle, Angle(0.0), Angle(360.0))
    assert result.value == 10.0


# ==============================================================================
# Unit system integration tests
# ==============================================================================


def test_custom_unit_system() -> None:
    """Test creating a custom unit system type."""

    class CustomUnitSystem:
        def __init__(self, length: str, time: str, mass: str) -> None:
            self.length = length
            self.time = time
            self.mass = mass

        def get_unit(self, dimension: str) -> str:
            return getattr(self, dimension, None)

    class CustomQuantity:
        def __init__(self, value: float, unit: str) -> None:
            self.value = value
            self.unit = unit
            # Infer system (simplified)
            if unit in ("m", "s", "kg"):
                self._system = CustomUnitSystem("m", "s", "kg")
            elif unit in ("km", "s", "g"):
                self._system = CustomUnitSystem("km", "s", "g")
            else:
                self._system = None

    # Register dispatch
    @plum.dispatch
    def unitsystem_of(obj: CustomQuantity, /) -> CustomUnitSystem | None:
        return obj._system

    # Test
    q = CustomQuantity(100.0, "m")
    system = unitsystem_of(q)
    assert system is not None
    assert system.length == "m"
    assert system.time == "s"
    assert system.mass == "kg"


# ==============================================================================
# Complete package integration tests
# ==============================================================================


def test_minimal_physics_package() -> None:
    """Test a minimal physics package that integrates with unxt-api."""

    # Define the types
    class Unit:
        def __init__(self, symbol: str, dimension: str) -> None:
            self.symbol = symbol
            self.dimension = dimension

    class Quantity:
        def __init__(self, value: float, unit: Unit) -> None:
            self.value = value
            self.unit = unit

    # Register with unxt-api
    @plum.dispatch
    def unit(obj: Unit, /) -> Unit:
        """Identity for Unit."""
        return obj

    @plum.dispatch
    def unit_of(obj: Quantity, /) -> Unit:
        """Get unit from Quantity."""
        return obj.unit

    @plum.dispatch
    def dimension_of(obj: Unit, /) -> str:
        """Get dimension from Unit."""
        return obj.dimension

    @plum.dispatch
    def dimension_of(obj: Quantity, /) -> str:
        """Get dimension from Quantity."""
        return obj.unit.dimension

    @plum.dispatch
    def ustrip(obj: Quantity, /) -> float:
        """Strip units from Quantity."""
        return obj.value

    # Use the package
    meter = Unit("m", "length")
    distance = Quantity(100.0, meter)

    # Test all integrations
    assert unit(meter).symbol == "m"
    assert unit_of(distance).symbol == "m"
    assert dimension_of(meter) == "length"
    assert dimension_of(distance) == "length"
    assert ustrip(distance) == 100.0


# ==============================================================================
# Dispatch priority tests
# ==============================================================================


def test_more_specific_dispatch_wins() -> None:
    """Test that more specific dispatches take priority."""

    class BaseType:
        def __init__(self, value: Any) -> None:
            self.value = value

    class SpecificType(BaseType):
        pass

    # Register base dispatch
    @plum.dispatch
    def unit_of(obj: BaseType, /) -> str:
        return "base"

    # Register more specific dispatch
    @plum.dispatch
    def unit_of(obj: SpecificType, /) -> str:
        return "specific"

    # Test
    base_obj = BaseType(42)
    specific_obj = SpecificType(42)

    assert unit_of(base_obj) == "base"
    assert unit_of(specific_obj) == "specific"


def test_multiple_implementations_coexist() -> None:
    """Test that multiple implementations can coexist."""

    class TypeA:
        pass

    class TypeB:
        pass

    @plum.dispatch
    def dimension(obj: TypeA, /) -> str:
        return "dimension_a"

    @plum.dispatch
    def dimension(obj: TypeB, /) -> str:
        return "dimension_b"

    a = TypeA()
    b = TypeB()

    assert dimension(a) == "dimension_a"
    assert dimension(b) == "dimension_b"


# ==============================================================================
# Error handling tests
# ==============================================================================


def test_no_matching_dispatch() -> None:
    """Test that missing dispatches raise appropriate errors."""

    class UnknownType:
        pass

    obj = UnknownType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.dimension(obj)


def test_custom_error_in_implementation() -> None:
    """Test that custom errors in implementations are propagated."""

    class ErrorType:
        pass

    @plum.dispatch
    def unit_of(obj: ErrorType, /) -> str:
        msg = "Custom error in implementation"
        raise ValueError(msg)

    obj = ErrorType()

    with pytest.raises(ValueError, match="Custom error in implementation"):
        unit_of(obj)



"""Unit tests for units API (unxt_api._units)."""

import plum
import pytest

import unxt_api as uapi

# ==============================================================================
# Tests for unit()
# ==============================================================================


def test_unit_is_abstract_dispatch() -> None:
    """Test that unit is an abstract dispatch function."""
    assert isinstance(uapi.unit, plum.function.Function)
    assert hasattr(uapi.unit, "methods")


def test_unit_accepts_any_type() -> None:
    """Test that unit accepts Any type in signature."""
    assert "unit" in dir(uapi)


def test_unit_no_default_implementation() -> None:
    """Test that calling unit without implementation raises error."""

    class NoDispatchType:
        pass

    obj = NoDispatchType()

    with pytest.raises(plum.resolver.NotFoundLookupError):
        uapi.unit(obj)


def test_unit_can_register_custom_dispatch(custom_unit_type) -> None:
    """Test that custom dispatches can be registered."""

    @plum.dispatch
    def unit(obj: custom_unit_type, /) -> str:
        return obj.unit_str

    obj = custom_unit_type("m")
    result = unit(obj)
    assert result == "m"


def test_unit_multiple_dispatches_possible() -> None:
    """Test that multiple dispatches can coexist."""
    initial_count = len(uapi.unit.methods)
    assert initial_count >= 0


# ==============================================================================
# Tests for unit_of()
# ==============================================================================


def test_unit_of_is_abstract_dispatch() -> None:
    """Test that unit_of is an abstract dispatch function."""
    assert isinstance(uapi.unit_of, plum.function.Function)
    assert hasattr(uapi.unit_of, "methods")


def test_unit_of_accepts_any_type() -> None:
    """Test that unit_of accepts Any type in signature."""
    assert "unit_of" in dir(uapi)


def test_unit_of_no_default_implementation() -> None:
    """Test that calling unit_of without implementation may return None.

    Note: When unxt is imported (e.g., during doctest execution), it may
    register a default implementation that returns None for unknown types.
    """

    class NoDispatchType:
        pass

    obj = NoDispatchType()

    try:
        result = uapi.unit_of(obj)
        # If it doesn't raise, it should return None (unxt's default)
        assert result is None
    except plum.resolver.NotFoundLookupError:
        # This is also acceptable if unxt hasn't been imported
        pass


def test_unit_of_can_register_custom_dispatch(custom_quantity_type) -> None:
    """Test that custom dispatches can be registered for unit_of."""

    @plum.dispatch
    def unit_of(obj: custom_quantity_type, /) -> str:
        return obj.unit_str

    obj = custom_quantity_type(42, "kg")
    result = unit_of(obj)
    assert result == "kg"


# ==============================================================================
# API consistency tests
# ==============================================================================


def test_both_unit_functions_are_exported() -> None:
    """Test that both functions are exported from unxt_api."""
    assert hasattr(uapi, "unit")
    assert hasattr(uapi, "unit_of")


def test_both_unit_functions_in_all() -> None:
    """Test that both functions are in __all__."""
    assert "unit" in uapi.__all__
    assert "unit_of" in uapi.__all__


def test_unit_functions_have_independent_registries() -> None:
    """Test that unit and unit_of have independent dispatch registries."""
    assert uapi.unit is not uapi.unit_of
    assert id(uapi.unit.methods) != id(uapi.unit_of.methods)



#!/usr/bin/env -S uv run --script  # noqa: EXE001
# /// script
#    dependencies = ["nox", "nox_uv"]
# ///
"""Nox setup."""

import argparse
import shutil
from enum import StrEnum, auto
from pathlib import Path
from typing import assert_never

import nox
from nox_uv import session

nox.needs_version = ">=2024.3.2"
nox.options.default_venv_backend = "uv"

DIR = Path(__file__).parent.resolve()


class PackageEnum(StrEnum):
    """Enum for package names."""

    @staticmethod
    def _generate_next_value_(name: str, *_: object, **__: object) -> str:
        return name

    def __repr__(self) -> str:
        return f"{self.value!r}"

    unxt = auto()
    hypothesis = auto()


# =============================================================================
# Comprehensive sessions


@session(
    uv_groups=["lint", "test", "docs"],
    uv_extras=["all"],
    reuse_venv=True,
    default=True,
)
def all(s: nox.Session, /) -> None:  # noqa: A001
    """Run all default sessions."""
    s.notify("lint")
    s.notify("test")
    s.notify("docs")


# =============================================================================
# Linting


@session(uv_groups=["lint"], reuse_venv=True)
def lint(s: nox.Session, /) -> None:
    """Run the linter."""
    s.notify("precommit")
    s.notify("pylint(package='unxt')")
    s.notify("pylint(package='hypothesis')")


@session(uv_groups=["lint"], reuse_venv=True)
def precommit(s: nox.Session, /) -> None:
    """Run pre-commit."""
    s.run("pre-commit", "run", "--all-files", *s.posargs)


@session(uv_groups=["lint"], reuse_venv=True)
@nox.parametrize("package", list(PackageEnum))
def pylint(s: nox.Session, /, package: PackageEnum) -> None:
    """Run PyLint."""
    match package:
        case PackageEnum.unxt:
            package_paths = ["packages/unxt-api/src", "src/unxt"]
        case PackageEnum.hypothesis:
            package_paths = ["packages/unxt-hypothesis/src"]
            s.install("-e", "packages/unxt-hypothesis")
        case _:
            assert_never(package)

    s.run("pylint", *package_paths, *s.posargs)


# =============================================================================
# Testing


@session(uv_groups=["test"], reuse_venv=True)
def test(s: nox.Session, /) -> None:
    """Run the unit and regular tests."""
    s.notify("pytest(package='unxt')", posargs=s.posargs)
    s.notify("pytest(package='hypothesis')", posargs=s.posargs)
    # s.notify("pytest_benchmark", posargs=s.posargs)


def _parse_pytest_paths(package: PackageEnum, /) -> list[str]:
    match package:
        case PackageEnum.unxt:
            package_paths = [
                "README.md",
                "docs",
                "src/",
                "tests/",
                "packages/unxt-api/",
            ]
        case PackageEnum.hypothesis:
            package_paths = ["packages/unxt-hypothesis/"]
        case _:
            assert_never(package)

    return package_paths


@session(uv_groups=["test"], uv_extras=["workspace"], reuse_venv=True)
@nox.parametrize("package", list(PackageEnum))
def pytest(s: nox.Session, /, package: PackageEnum) -> None:
    """Run the unit and regular tests."""
    package_paths = _parse_pytest_paths(package)
    s.run("pytest", *package_paths, *s.posargs)


@session(uv_groups=["test-all"], uv_extras=["interop-mpl"], reuse_venv=True)
@nox.parametrize("package", list(PackageEnum))
def pytest_all(s: nox.Session, /, package: PackageEnum) -> None:
    """Run the unit and regular tests."""
    package_paths = _parse_pytest_paths(package)
    s.run("pytest", *package_paths, *s.posargs)


@session(uv_groups=["test"], reuse_venv=True)
def pytest_benchmark(s: nox.Session, /) -> None:
    """Run the benchmarks."""
    s.run("pytest", "tests/benchmark", "--codspeed", *s.posargs)


# =============================================================================
# Documentation


@session(uv_groups=["docs"], uv_extras=["workspace"], reuse_venv=True)
def docs(s: nox.Session, /) -> None:
    """Build the docs. Pass "--serve" to serve. Pass "-b linkcheck" to check links."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--serve", action="store_true", help="Serve after building")
    parser.add_argument(
        "-b", dest="builder", default="html", help="Build target (default: html)"
    )
    parser.add_argument("--offline", action="store_true", help="run in offline mode")
    parser.add_argument("--output-dir", dest="output_dir", default="_build")
    args, posargs = parser.parse_known_args(s.posargs)

    if args.builder != "html" and args.serve:
        s.error("Must not specify non-HTML builder with --serve")

    s.chdir("docs")

    # Generate custom intersphinx inventories
    s.run("python", "_static/generate_jaxtyping_inv.py")
    s.run("python", "_static/generate_equinox_inv.py")
    s.run("python", "_static/generate_quax_blocks_inv.py")

    if args.builder == "linkcheck":
        s.run("sphinx-build", "-b", "linkcheck", ".", "_build/linkcheck", *posargs)
        return

    shared_args = (
        "-n",  # nitpicky mode
        "-T",  # full tracebacks
        f"-b={args.builder}",
        f"-d={args.output_dir}/doctrees",
        "-D",
        "language=en",
        ".",
        f"{args.output_dir}/{args.builder}",
        *posargs,
    )

    if args.serve:
        s.run("sphinx-autobuild", *shared_args)
    else:
        s.run("sphinx-build", "--keep-going", *shared_args)


@session(uv_groups=["docs"], reuse_venv=True)
def build_api_docs(s: nox.Session, /) -> None:
    """Build (regenerate) API docs."""
    s.chdir("docs")
    s.run(
        "sphinx-apidoc",
        "-o",
        "api/",
        "--module-first",
        "--no-toc",
        "--force",
        "../src/unxt",
    )


# =============================================================================
# Packaging


@session(uv_groups=["build"])
def build(s: nox.Session, /) -> None:
    """Build an SDist and wheel."""
    build_path = DIR.joinpath("build")
    if build_path.exists():
        shutil.rmtree(build_path)

    s.run("python", "-m", "build")


# =============================================================================

if __name__ == "__main__":
    nox.main()



"""Angular quantities."""

__all__ = ("Angle",)

from typing import final

import equinox as eqx
from jaxtyping import Array, Real

import unxt_api as uapi
from .base_angle import AbstractAngle
from .value import StaticValue, convert_to_quantity_value
from unxt.units import AbstractUnit


@final
class Angle(AbstractAngle):
    """Angular quantity.

    Examples
    --------
    >>> import unxt as u

    Create an Angle:

    >>> q = u.Angle(1, "rad")
    >>> q
    Angle(Array(1, dtype=int32, ...), unit='rad')

    Wrap an Angle to a range:

    >>> q = u.Angle(370, "deg")
    >>> q.wrap_to(u.Q(0, "deg"), u.Q(360, "deg"))
    Angle(Array(10, dtype=int32, ...), unit='deg')

    Create an Angle array:

    >>> q = u.Angle([1, 2, 3], "deg")
    >>> q
    Angle(Array([1, 2, 3], dtype=int32), unit='deg')

    Do math on an Angle:

    >>> 2 * q
    Angle(Array([2, 4, 6], dtype=int32), unit='deg')

    >>> q % u.Q(4, "deg")
    Angle(Array([1, 2, 3], dtype=int32), unit='deg')

    """

    value: Real[Array | StaticValue, "*shape"] = eqx.field(
        converter=convert_to_quantity_value
    )
    """The value of the `unxt.AbstractQuantity`."""

    unit: AbstractUnit = eqx.field(static=True, converter=uapi.unit)
    """The unit associated with this value."""



"""Hypothesis strategies for Quantity objects."""

__all__ = ("quantities", "wrap_to", "angles")

from typing import Any, T, TypeAlias, TypeVar

import jax.numpy as jnp
import numpy as np
from hypothesis import strategies as st
from hypothesis.extra.array_api import make_strategies_namespace

import unxt as u
from .units import units

T = TypeVar("T")

# Create array API strategies namespace for JAX
xps = make_strategies_namespace(jnp)


def draw_if_strategy(draw: st.DrawFn, v: T | st.SearchStrategy[T], /) -> T:
    """Draw a value if a strategy is given, else return the value."""
    return draw(v) if isinstance(v, st.SearchStrategy) else v


@st.composite
def quantities(
    draw: st.DrawFn,
    /,
    unit: str
    | u.AbstractUnit
    | st.SearchStrategy[u.AbstractUnit | u.AbstractDimension]
    | u.AbstractDimension = units(),  # noqa: B008
    *,
    quantity_cls: type[u.AbstractQuantity] = u.Quantity,
    dtype: Any | st.SearchStrategy[np.dtype] = jnp.float32,
    shape: int | tuple[int, ...] | st.SearchStrategy[int | tuple[int, ...]] = (),
    elements: st.SearchStrategy[float] | None = None,
    unique: bool = False,
    static_value: bool | st.SearchStrategy[bool] = False,
    **kwargs: Any,
) -> u.AbstractQuantity:
    """Generate hypothesis strategy for unxt Quantity objects.

    This strategy combines JAX array generation with unit specifications to
    create valid Quantity objects for property-based testing.

    Parameters
    ----------
    draw : st.DrawFn
        Hypothesis draw function (automatically provided by @st.composite).
    unit : str | st.SearchStrategy[unxt.AbstractUnit]
        Unit specification for the Quantity. Can be:

        - str: Fixed unit (e.g., "kpc", "km/s")
        - `unxt.AbstractUnit`: Fixed unit object
        - `unxt.AbstractDimension`: Dimension to derive unit from (uses
          `unxt_hypothesis.units` strategy)
        - SearchStrategy: Strategy that generates units (e.g., from `units()`)
          or dimensions. The default strategy samples from SI base units.
    quantity_cls : type[`unxt.AbstractQuantity`], optional
        The target quantity class to convert to. Default is `unxt.Quantity`.
        Can be any `unxt.AbstractQuantity` subclass like `unxt.Quantity` or
        `unxt.Angle`.
    dtype : Any, optional
        NumPy/JAX dtype for the underlying array. Default is jnp.float32.
        Can also be a `SearchStrategy` that generates dtypes.
    shape : int | tuple[int, ...] | st.SearchStrategy[tuple[int, ...]]
        Shape of the array. Can be:

        - int: 1D array of that length
        - tuple: fixed shape
        - `SearchStrategy`: strategy that generates shapes
    elements : st.SearchStrategy[float] | None, optional
        Strategy for generating array elements. If `None`, uses finite floats.
    unique : bool, optional
        Whether array elements should be unique. Default is `False`.
    static_value : bool, optional
        Whether to wrap the generated array in `unxt.quantity.StaticValue`.
        Default is `False`.
    **kwargs : Any
        Additional keyword arguments (currently unused, reserved for future
        use).

    Returns
    -------
    `unxt.Quantity`
        A Quantity object with the specified unit and array properties.

    Examples
    --------
    Basic usage with fixed unit and shape:

    >>> from hypothesis import given, strategies as st
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import unxt_hypothesis as ust

    >>> @given(q=ust.quantities("kpc", shape=3))
    ... def test_position(q):
    ...     assert q.unit == u.unit("kpc")
    ...     assert q.shape == (3,)

    >>> @given(q=ust.quantities("km/s", shape=()))
    ... def test_scalar_velocity(q):
    ...     assert q.unit == u.unit("km/s")
    ...     assert q.shape == ()

    Using a unit strategy to vary units across test runs:

    >>> @given(q=ust.quantities(ust.units("length"), shape=3))
    ... def test_length_with_varying_units(q):
    ...     # Unit will vary across test runs
    ...     assert u.dimension_of(q) == u.dimension("length")
    ...     assert q.shape == (3,)

    As a convenience, instead of nesting strategies with ``ust.units``, you can
    directly pass the dimension object (not a string) to ``ust.quantities``:

    >>> @given(q=ust.quantities(u.dimension("velocity"), shape=3))
    ... def test_velocity_from_dimension(q):
    ...     # Will generate different velocity units (m/s, km/s, etc.)
    ...     assert u.dimension_of(q) == u.dimension("velocity")
    ...     assert q.shape == (3,)

    This (and ``ust.units``) support a strategy for dimensions:

    >>> dim_strat = st.sampled_from([u.dimension("length"), u.dimension("mass")])
    >>> @given(q=ust.quantities(dim_strat, shape=()))
    ... def test_mixed_dimensions(q):
    ...     # Will generate either length or mass quantities
    ...     dim = u.dimension_of(q)
    ...     assert dim in (u.dimension("length"), u.dimension("mass"))

    Using dtype as a strategy:

    >>> dtype_strat = st.sampled_from([jnp.float32, jnp.float64])
    >>> @given(q=ust.quantities("m", dtype=dtype_strat))
    ... def test_varying_dtype(q):
    ...     # dtype will vary between float32 and float64
    ...     assert q.dtype in (jnp.float32, jnp.float64)

    Using elements to constrain value ranges for distances (positive values):

    >>> @given(
    ...     q=ust.quantities(
    ...         "kpc", shape=3, elements=st.floats(min_value=0, max_value=100, width=32)
    ...     )
    ... )
    ... def test_positive_distance(q):
    ...     # All elements will be positive (suitable for distances)
    ...     assert jnp.all(q.value >= 0)
    ...     assert jnp.all(q.value <= 100)

    Using elements for longitude angles (0 to 360 degrees):

    >>> @given(
    ...     q=ust.quantities(
    ...         "deg",
    ...         shape=(),
    ...         elements=st.floats(min_value=0, max_value=360, width=32),
    ...     )
    ... )
    ... def test_longitude_range(q):
    ...     # Longitude in [0, 360] degrees
    ...     assert 0 <= q.value <= 360

    Using quantity_cls to generate Angle objects:

    >>> @given(angle=ust.quantities("rad", quantity_cls=u.Angle, shape=3))
    ... def test_angle_generation(angle):
    ...     # Creates Angle instances instead of Quantity
    ...     assert isinstance(angle, u.Angle)
    ...     assert angle.unit == u.unit("rad")
    ...     assert angle.shape == (3,)

    """
    # Handle unit specification - draw from strategy if needed
    unit = draw_if_strategy(draw, unit)
    unit_obj = (
        draw(units(unit)) if isinstance(unit, u.AbstractDimension) else u.unit(unit)
    )

    # Handle shape specification
    shape = draw_if_strategy(draw, shape)
    array_shape = (shape,) if isinstance(shape, int) else shape

    # DType handling
    dtype = draw_if_strategy(draw, dtype)
    static_value = draw_if_strategy(draw, static_value)

    # Default elements strategy if not provided
    if elements is None:
        # Use xps.from_dtype to get the appropriate strategy for the dtype
        elements = xps.from_dtype(dtype, allow_nan=False, allow_infinity=False)

    # Generate array using JAX array-api strategy
    # For arrays, use the JAX array API strategies
    values = draw(
        xps.arrays(dtype=dtype, shape=array_shape, elements=elements, unique=unique)
    )
    if static_value:
        values = u.quantity.StaticValue(np.asarray(values))

    # Create Quantity with the specified unit
    return quantity_cls(values, unit_obj, **kwargs)


@st.composite
def wrap_to(
    draw: st.DrawFn,
    quantity: u.AbstractQuantity | st.SearchStrategy[u.AbstractQuantity],
    /,
    min: u.AbstractQuantity | st.SearchStrategy[u.AbstractQuantity],
    max: u.AbstractQuantity | st.SearchStrategy[u.AbstractQuantity],
) -> u.AbstractQuantity:
    """Generate hypothesis strategy for wrapped quantities.

    This strategy takes a quantity strategy and wraps the generated values
    to the specified [min, max) range using the wrap_to function.

    Parameters
    ----------
    draw
        Hypothesis draw function (automatically provided by @st.composite).
    quantity
        Quantity or strategy that generates the base quantity to wrap.
    min
        Minimum value of the wrapping range (inclusive).
    max
        Maximum value of the wrapping range (exclusive).

    Returns
    -------
    unxt.AbstractQuantity
        The wrapped quantity in the range [min, max).

    Examples
    --------
    Wrap angles to 0-360 degree range:

    >>> from hypothesis import given
    >>> import hypothesis.strategies as st
    >>> import unxt as u
    >>> import unxt_hypothesis as ust

    >>> @given(
    ...     angle=ust.wrap_to(
    ...         ust.quantities("deg", quantity_cls=u.Angle),
    ...         min=u.Q(0, "deg"),
    ...         max=u.Q(360, "deg"),
    ...     )
    ... )
    ... def test_wrapped_angle(angle):
    ...     assert 0 <= angle.value <= 360

    Wrap angles with dynamic min/max:

    >>> @given(
    ...     angle=ust.wrap_to(
    ...         ust.quantities("rad", quantity_cls=u.Angle),
    ...         min=st.just(u.Q(0, "rad")),
    ...         max=st.just(u.Q(6.28318530718, "rad")),
    ...     )
    ... )
    ... def test_wrapped_angle_rad(angle):
    ...     assert 0 <= angle.value <= 6.28318530718

    """
    # Wrap the quantity to the specified range
    return u.quantity.wrap_to(
        draw_if_strategy(draw, quantity),
        draw_if_strategy(draw, min),
        draw_if_strategy(draw, max),
    )


# Alias for use in angles strategy
_wrap_to_strategy = wrap_to

WrapToArgs: TypeAlias = (
    tuple[u.Q | st.SearchStrategy[u.Q], u.Q | st.SearchStrategy[u.Q]] | None
)


@st.composite  # type: ignore[untyped-decorator]
def angles(
    draw: st.DrawFn,
    *,
    wrap_to: WrapToArgs | st.SearchStrategy[WrapToArgs] = None,  # pylint: disable=redefined-outer-name
    **kwargs: Any,
) -> u.Angle:
    """Strategy for generating Angle instances.

    Parameters
    ----------
    draw
        Hypothesis draw function. Automatically provided by hypothesis.
    wrap_to
        Strategy for generating (min, max) bounds for angle wrapping.
        If `None`, the angle will have no wrapping specified.
    **kwargs
        Additional keyword arguments passed to `unxt_hypothesis.quantities`.
        Common options include 'dtype', 'shape', 'elements', 'unique'.  The
        arguments 'unit' and 'quantity_cls' are set automatically and should not
        be provided.

    Returns
    -------
    unxt.Angle
        A strategy that generates Angle instances.

    Examples
    --------
    >>> from hypothesis import given
    >>> from unxt_hypothesis import angles
    >>> import unxt as u

    >>> @given(angle=angles())
    ... def test_angle(angle):
    ...     assert isinstance(angle, u.Angle)

    With wrapping:

    >>> from hypothesis import strategies as st
    >>> wrap_bounds = st.just((u.Q(0, "deg"), u.Q(360, "deg")))
    >>> @given(angle=angles(wrap_to=wrap_bounds))
    ... def test_wrapped_angle(angle):
    ...     assert angle.wrap_to is not None

    """
    # Determine wrapping bounds if provided
    wrap_to = draw_if_strategy(draw, wrap_to)

    # Extract unit if provided (to avoid conflicts with dimension) Default to
    # angle dimension, but user can override with specific unit. If user
    # provides a bad unit, unxt_hypothesis.quantities will raise an error given
    # the `u.Angle` quantity_cls.
    unit = kwargs.pop("unit", u.dimension("angle"))

    # Generate the Angle quantity
    angle_strategy = quantities(unit, quantity_cls=u.Angle, **kwargs)

    # Apply wrapping if bounds are provided
    if wrap_to is not None:
        wrap_min, wrap_max = wrap_to
        return draw(_wrap_to_strategy(angle_strategy, min=wrap_min, max=wrap_max))

    return draw(angle_strategy)


# Register type strategy for Hypothesis's st.from_type()
# Note: Pass the callable, not an invoked strategy
st.register_type_strategy(
    u.AbstractQuantity, lambda _: quantities(quantity_cls=u.Quantity)
)
st.register_type_strategy(u.Quantity, lambda typ: quantities(quantity_cls=typ))
st.register_type_strategy(
    u.quantity.BareQuantity, lambda typ: quantities(quantity_cls=typ)
)
st.register_type_strategy(
    u.quantity.StaticQuantity,
    lambda typ: quantities(quantity_cls=typ, static_value=True),
)
st.register_type_strategy(u.Angle, lambda _: angles())



"""Dimensions compatibility."""

__all__: tuple[str, ...] = ()


from dataclasses import MISSING, Field

import astropy.units as apyu
import plum

from unxt.dims import AbstractDimension


@plum.dispatch
def dimension_of(obj: apyu.Quantity, /) -> AbstractDimension:
    """Return the dimension of a quantity.

    Examples
    --------
    >>> import unxt as u
    >>> import astropy.units as apyu

    >>> q = apyu.Quantity(1, "m")
    >>> u.dimension_of(q)
    PhysicalType('length')

    """
    return dimension_of(obj.unit)


# ===================================================
# `Dataclassish` support


@plum.dispatch
def fields(obj: AbstractDimension, /) -> tuple[Field, ...]:
    """Return the fields of a dimension.

    Examples
    --------
    >>> import dataclassish as dc
    >>> import astropy.units as apyu

    >>> dim = apyu.get_physical_type("length")
    >>> dc.fields(dim)
    (Field(name='_unit',...), Field(name='_physical_type',...))

    """
    unit_field = Field(MISSING, MISSING, True, False, False, False, {}, False)  # noqa: FBT003
    unit_field.name = "_unit"
    physical_type_field = Field(MISSING, MISSING, True, False, False, False, {}, False)  # noqa: FBT003
    physical_type_field.name = "_physical_type"
    return (unit_field, physical_type_field)



"""Tests for the angles strategy."""

import hypothesis.strategies as st
from hypothesis import given, settings

import unxt as u
from unxt_hypothesis import angles


@given(angle=angles())
@settings(max_examples=50)
def test_basic_angle(angle: u.Angle) -> None:
    """Test basic angle generation."""
    assert isinstance(angle, u.Angle)
    assert angle.shape == ()


@given(angle=angles(unit="deg", elements={"min_value": -180, "max_value": 180}))
@settings(max_examples=50)
def test_angle_with_bounds(angle: u.Angle) -> None:
    """Test angle generation with value bounds."""
    assert isinstance(angle, u.Angle)
    assert -180 <= angle.to("deg").value <= 180


@given(angle=angles(unit="rad"))
@settings(max_examples=50)
def test_angle_radians(angle: u.Angle) -> None:
    """Test angle generation in radians."""
    assert isinstance(angle, u.Angle)
    assert angle.unit == "rad"


@given(angle=angles(wrap_to=st.just((u.Q(0, "deg"), u.Q(360, "deg")))))
@settings(max_examples=50)
def test_angle_with_wrapping(angle: u.Angle) -> None:
    """Test angle generation with wrapping."""
    assert isinstance(angle, u.Angle)
    # The wrap_to strategy wraps values to [min, max) range
    # but floating point can give us exactly max
    angle_deg = angle.to("deg").value
    assert 0 <= angle_deg <= 360


@given(angle=angles(wrap_to=st.just((u.Q(-180, "deg"), u.Q(180, "deg")))))
@settings(max_examples=50)
def test_angle_with_symmetric_wrapping(angle: u.Angle) -> None:
    """Test angle generation with symmetric wrapping."""
    assert isinstance(angle, u.Angle)
    # The wrap_to strategy wraps values to [min, max) range
    # but floating point can give us exactly max
    angle_deg = angle.to("deg").value
    assert -180 <= angle_deg <= 180


@given(angle=angles(shape=5))
@settings(max_examples=30)
def test_angle_vector(angle: u.Angle) -> None:
    """Test vector angle generation."""
    assert isinstance(angle, u.Angle)
    assert angle.shape == (5,)


@given(angle=angles(shape=(2, 3)))
@settings(max_examples=30)
def test_angle_2d(angle: u.Angle) -> None:
    """Test 2D angle array generation."""
    assert isinstance(angle, u.Angle)
    assert angle.shape == (2, 3)


@given(angle=angles(shape=st.tuples(st.integers(1, 5), st.integers(1, 5))))
@settings(max_examples=30)
def test_angle_dynamic_shape(angle: u.Angle) -> None:
    """Test angle generation with dynamic shape."""
    assert isinstance(angle, u.Angle)
    assert len(angle.shape) == 2
    assert 1 <= angle.shape[0] <= 5
    assert 1 <= angle.shape[1] <= 5



"""Tests for the unitsystems strategy."""
# pylint: disable=unreachable

from hypothesis import given, settings

import unxt as u
import unxt_hypothesis as ust


@given(usys=ust.unitsystems("m", "s", "kg", "rad"))
@settings(max_examples=30)
def test_mks_system(usys: u.AbstractUnitSystem) -> None:
    """Test MKS unit system generation."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    assert usys["length"] == u.unit("m")
    assert usys["time"] == u.unit("s")
    assert usys["mass"] == u.unit("kg")
    assert usys["angle"] == u.unit("rad")


@given(usys=ust.unitsystems("kpc", "Myr", "Msun", "rad"))
@settings(max_examples=30)
def test_galactic_system(usys: u.AbstractUnitSystem) -> None:
    """Test galactic unit system generation."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    assert u.dimension_of(usys["length"]) == u.dimension("length")
    assert u.dimension_of(usys["time"]) == u.dimension("time")
    assert u.dimension_of(usys["mass"]) == u.dimension("mass")


@given(usys=ust.unitsystems(ust.units("length"), "s", "kg", "rad"))
@settings(max_examples=50)
def test_varying_length_unit(usys: u.AbstractUnitSystem) -> None:
    """Test unit system with varying length units."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    # Length varies, but should always have length dimension
    assert u.dimension_of(usys["length"]) == u.dimension("length")
    # Other units are fixed
    assert usys["time"] == u.unit("s")
    assert usys["mass"] == u.unit("kg")
    assert usys["angle"] == u.unit("rad")


@given(usys=ust.unitsystems(ust.units("length"), ust.units("time"), "kg", "rad"))
@settings(max_examples=50)
def test_varying_length_and_time(usys: u.AbstractUnitSystem) -> None:
    """Test unit system with varying length and time units."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    # Length and time vary
    assert u.dimension_of(usys["length"]) == u.dimension("length")
    assert u.dimension_of(usys["time"]) == u.dimension("time")
    # Mass and angle are fixed
    assert usys["mass"] == u.unit("kg")
    assert usys["angle"] == u.unit("rad")


@given(
    usys=ust.unitsystems(
        ust.units("length"), ust.units("time"), ust.units("mass"), "rad"
    )
)
@settings(max_examples=50)
def test_three_varying_units(usys: u.AbstractUnitSystem) -> None:
    """Test unit system with three varying units."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    assert u.dimension_of(usys["length"]) == u.dimension("length")
    assert u.dimension_of(usys["time"]) == u.dimension("time")
    assert u.dimension_of(usys["mass"]) == u.dimension("mass")
    assert usys["angle"] == u.unit("rad")


@given(
    usys=ust.unitsystems(
        ust.units("length", max_complexity=3),
        ust.units("time", max_complexity=3),
        "Msun",
        "rad",
    )
)
@settings(max_examples=30)
def test_complex_varying_units(usys: u.AbstractUnitSystem) -> None:
    """Test unit system with complex varying units."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    assert u.dimension_of(usys["length"]) == u.dimension("length")
    assert u.dimension_of(usys["time"]) == u.dimension("time")


@given(usys=ust.unitsystems("m", "s", "kg", "rad"))
@settings(max_examples=30)
def test_four_unit_system(usys: u.AbstractUnitSystem) -> None:
    """Test system with four units."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    assert u.dimension_of(usys["length"]) == u.dimension("length")
    assert u.dimension_of(usys["time"]) == u.dimension("time")


def test_accepts_fixed_string() -> None:
    """Test that the strategy accepts fixed string units."""
    strategy = ust.unitsystems("m", "s", "kg")
    assert strategy is not None


def test_accepts_unit_strategy() -> None:
    """Test that the strategy accepts unit strategies."""
    strategy = ust.unitsystems(ust.units("length"), ust.units("time"), "kg")
    assert strategy is not None


@given(usys=ust.unitsystems("AU", "yr", "Msun", "rad"))
@settings(max_examples=30)
def test_astronomical_units(usys: u.AbstractUnitSystem) -> None:
    """Test unit system with astronomical units."""
    assert isinstance(usys, u.AbstractUnitSystem)
    assert len(usys) == 4
    assert usys["length"] == u.unit("AU")
    assert usys["time"] == u.unit("yr")
    assert usys["mass"] == u.unit("Msun")
